' World-o-Mat Math-Edition 1
' Rune A.J.M, June 2007
' <runemoennike@gmail.com>
' For APIStudios, Gods and Idols
'
'
'
' Version 1b heavily modified by
' J. Pihl, March 5th 2008
' <jpihl@apiservers.com>
'
' Changes include: Removal of all lookup tables
' This makes getHeight slower, but it is now capable of generating infinite terrain
' No more resetNow since the lookup tables are gone
' Function clutter cleaned up
'
'
'
'
' Generates a heightmap using various noise/plasma/random techniques.
' Can generate for any given point without needing info about any other part.
' Calling the getHeight with the same coordinate will return the same height
'---------
' The higher these details levels are, the more fractured the map will be
' Level1 is the most significant one, a low value (3-5) will generate a map with
' only one or a few huge blobs of land, where as a higher value will lead to more
' "continents" being formed. Level 4 is the least significant one.
' Upping level 1: More continents
' Upping level 2: More "coast of norway"
' Upping level 3: Dunno.
' Upping level 4: More "sandy" look, more realistic for a "zoomed-out" view.
' Where "upping" is for example 5-10 times as much as default. (default: 10, 20, 40, 80)
' Play around with these! (Try 20, 80, 200, 400)
#Include "crt/math.bi"

dim shared as integer LEVEL1_DETAIL, LEVEL2_DETAIL, LEVEL3_DETAIL, LEVEL4_DETAIL
LEVEL1_DETAIL=20
LEVEL2_DETAIL=80
LEVEL3_DETAIL=160
LEVEL4_DETAIL=320
declare function getHeight(byval x as integer, byval y as integer) as double
declare function interpolate(a as double, b as double, x as double) as double
declare function noise(x as integer, y as integer) as double
declare function SmoothedNoise(x as integer, y as integer) as double
declare function InterpolatedNoise1(x as integer, y as integer) as double
declare function InterpolatedNoise2(x as integer, y as integer) as double
declare function InterpolatedNoise3(x as integer, y as integer) as double
declare function InterpolatedNoise4(x as integer, y as integer) as double

dim shared as integer fw, fh, NumOfSegments=128
dim shared as single randMul(2047)
'Dim Shared m_PI As Single =3.14
Dim Shared PI As Single =M_PI
Dim Shared 	numTriangles As UInteger
'+=numPolys

fw = 256
fh = 256

Sub drawSphere()
	Static As GLuint listnum = 0, numPolys
	dim as integer NumOfPoints = (NumOfSegments + 1) * (NumOfSegments + 1)
	Redim As glfloat points(NumOfPoints*3-1)
	dim as integer yc, xc, i
	dim as single noise__
	
	If (listnum=0) Then
		Dim As Single  UR,YP,VR,UW,VW,l
		Dim As Single  US = (m_PI*2) / NumOfSegments
		Dim As Single  VS =  m_PI    / NumOfSegments
		Dim As Integer PC =0
		For yc = 0 To NumOfSegments
			UR = Sin(VW)
			YP = Cos(VW)
			VR = Sin(VW)
			VW+=VS
			
			UW = 0
			For xc = 0 To NumOfSegments
				noise__ = getHeight((xc-NumOfSegments/2), yc+2768)
				if noise__ > 1 then noise__ = 1
				if noise__ < 0 then noise__ = 0
				Points(PC*3+0)=Sin(UW) * UR * 			(.99 + noise__/1)
				Points(PC*3+1)=YP *						(.99 + noise__/1)
				Points(PC*3+2)=Cos(m_PI + UW) * VR *	(.99 + noise__/1)
				PC+=1: UW+=US
			Next
		Next
		listnum = glGenLists(1)
		glNewList (listnum,GL_COMPILE)
		glBegin (GL_TRIANGLES)
		For yc = 0 To NumOfSegments - 1
			For xc = 0 To NumOfSegments - 1
				Dim As Integer P0 = (yc + 0) * (NumOfSegments + 1) + (xc + 0)
				Dim As Integer P1 = (yc + 0) * (NumOfSegments + 1) + (xc + 1)
				Dim As Integer P2 = (yc + 1) * (NumOfSegments + 1) + (xc + 1)
				Dim As Integer P3 = (yc + 1) * (NumOfSegments + 1) + (xc + 0)
				Dim As VECTOR_3D_T v, n
				v.x=Points(p0*3+0):v.y=Points(p0*3+1):v.z=Points(p0*3+2)
				n=v
				n=vector_normalize(n)
				glColor3fv(@v.x)
				glNormal3fv(@n.x)
				glTexCoord2f( asin(n.x)/PI+0.5f, asin(n.y)/PI+0.5f )
				glVertex3fv(@v.x)
				
				v.x=Points(p1*3+0):v.y=Points(p1*3+1):v.z=Points(p1*3+2)
				n=v
				n=vector_normalize(n)
				glNormal3fv(@n.x)
				glTexCoord2f( asin(n.x)/PI+0.5f, asin(n.y)/PI+0.5f )
				glVertex3fv(@v.x)
				
				v.x=Points(p3*3+0):v.y=Points(p3*3+1):v.z=Points(p3*3+2)
				n=v
				l=n.x*n.x + n.y*n.y +n.z*n.z
				If l<>0 Then
					l=1/Sqr(l):n.x*=l:n.y*=l:n.z*=l
				End If
				glNormal3fv(@n.x)
				glTexCoord2f( asin(n.x)/PI+0.5f, asin(n.y)/PI+0.5f )
				glVertex3fv(@v.x)
				
				v.x=Points(p1*3+0):v.y=Points(p1*3+1):v.z=Points(p1*3+2)
				n=v
				n=vector_normalize(n)
				glNormal3fv(@n.x)
				glTexCoord2f( asin(n.x)/PI+0.5f, asin(n.y)/PI+0.5f )
				glVertex3fv(@v.x)
				
				v.x=Points(p2*3+0):v.y=Points(p2*3+1):v.z=Points(p2*3+2)
				n=v
				n=vector_normalize(n)
				glNormal3fv(@n.x)
				glTexCoord2f( asin(n.x)/PI+0.5f, asin(n.y)/PI+0.5f )
				glVertex3fv(@v.x)
				
				v.x=Points(p3*3+0):v.y=Points(p3*3+1):v.z=Points(p3*3+2)
				n=v
				n=vector_normalize(n)
				glNormal3fv(@n.x)
				glTexCoord2f( asin(n.x)/PI+0.5f, asin(n.y)/PI+0.5f )
				glVertex3fv(@v.x)
				numPolys+=6
			Next
		Next
		glEnd()
		glEndList()
		
	End If
	glCallList(listnum)
	numTriangles+=numPolys
End Sub

function getHeight(byval x as integer, byval y as integer) as double
	dim as double h
	x=abs(x)
	y=abs(y)
	h = tan(InterpolatedNoise1(x, y) * 1.000 + _
			InterpolatedNoise2(x, y) * 0.500 + _
			InterpolatedNoise3(x, y) * 0.250 + _
			InterpolatedNoise4(x, y) * 0.125)
	h += 0.5
	h = 3 * h^2 - 2 * h^3
	return h
end function

function interpolate(a as double, b as double, x as double) as double        
	dim as double ft, f
	ft = x * PI
	f = (1 - cos(ft)) * .5
	return  a*(1-f) + b*f
end function

function noise(x as integer, y as integer) as double
	if x>-1 and y>-1 then
		if randMul(0)=0 then
			dim as integer rx
			randomize -5
			for rx=0 to 2047
				randMul(rx)=rnd
			next
		end if
		return randMul((1+x*y+sin(x)*128+cos(y)*128) mod 2048) * 2 - 1
	else
		return 0
	end if
end function

function SmoothedNoise(x as integer, y as integer) as double
	dim as double corners, sides, center
	corners = ( noise(x-1, y-1)+noise(x+1, y-1)+noise(x-1, y+1)+noise(x+1, y+1) ) / 16
	sides   = ( noise(x-1, y)  +noise(x+1, y)  +noise(x, y-1)  +noise(x, y+1) ) /  8
	center  =  noise(x, y) / 4
	return corners + sides + center
end function

function InterpolatedNoise4(x as integer, y as integer) as double
	dim as double lx, ly, rx, ry, px, py, h, i1, i2
	dim as integer nx, ny
	lx = x
	ly = y
	nx = int(lx / fw * LEVEL4_DETAIL)
	ny = int(ly / fh * LEVEL4_DETAIL)
	rx = fw / LEVEL4_DETAIL
	ry = fh / LEVEL4_DETAIL
	px = (lx - (nx * rx)) / rx
	py = (ly - (ny * ry)) / ry
	i1 = interpolate(SmoothedNoise(nx, ny), SmoothedNoise(nx + 1, ny), px)
	i2 = interpolate(SmoothedNoise(nx, ny + 1), SmoothedNoise(nx + 1, ny + 1), px)
	h = interpolate(i1, i2, py)
	return h
end function

function InterpolatedNoise3(x as integer, y as integer) as double
	dim as double lx, ly, rx, ry, px, py, h, i1, i2
	dim as integer nx, ny
	lx = x
	ly = y
	nx = int(lx / fw * LEVEL3_DETAIL)
	ny = int(ly / fh * LEVEL3_DETAIL)
	rx = fw / LEVEL3_DETAIL
	ry = fh / LEVEL3_DETAIL
	px = (lx - (nx * rx)) / rx
	py = (ly - (ny * ry)) / ry
	i1 = interpolate(SmoothedNoise(nx, ny), SmoothedNoise(nx + 1, ny), px)
	i2 = interpolate(SmoothedNoise(nx, ny + 1), SmoothedNoise(nx + 1, ny + 1), px)
	h = interpolate(i1, i2, py)
	return h
end function

function InterpolatedNoise2(x as integer, y as integer) as double
	dim as double lx, ly, rx, ry, px, py, h, i1, i2
	dim as integer nx, ny
	lx = x
	ly = y
	nx = int(lx / fw * LEVEL2_DETAIL)
	ny = int(ly / fh * LEVEL2_DETAIL)
	rx = fw / LEVEL2_DETAIL
	ry = fh / LEVEL2_DETAIL
	px = (lx - (nx * rx)) / rx
	py = (ly - (ny * ry)) / ry
	i1 = interpolate(SmoothedNoise(nx, ny), SmoothedNoise(nx + 1, ny), px)
	i2 = interpolate(SmoothedNoise(nx, ny + 1), SmoothedNoise(nx + 1, ny + 1), px)
	h = interpolate(i1, i2, py)
	return h
end function

function InterpolatedNoise1(x as integer, y as integer) as double
	dim as double lx, ly, rx, ry, px, py, h, i1, i2
	dim as integer nx, ny
	lx = x
	ly = y
	nx = int(lx / fw * LEVEL1_DETAIL)
	ny = int(ly / fh * LEVEL1_DETAIL)
	rx = fw / LEVEL1_DETAIL
	ry = fh / LEVEL1_DETAIL
	px = (lx - (nx * rx)) / rx
	py = (ly - (ny * ry)) / ry
	i1 = interpolate(SmoothedNoise(nx, ny), SmoothedNoise(nx + 1, ny), px)
	i2 = interpolate(SmoothedNoise(nx, ny + 1), SmoothedNoise(nx + 1, ny + 1), px)
	h = interpolate(i1, i2, py)
	return h
end function
