Declare Function GuiTextEditor(bounds As Rectangle, text As ZString ptr, textLen0 As integer, editMode As boolean) As boolean

' Text editor control (Advanced text box)
Function GuiTextEditor(bounds As Rectangle, text0 As ZString ptr, textLen0 As integer, editMode As boolean) As boolean

    static As Rectangle cursor     ' Cursor position and size
    static As Integer framesCounter = 0       ' Blinking cursor frames counter
    static As Integer cursorCodepoint = -1
    static As Integer selectStartCp = -1
    static As Integer selectLengthCp = 0
    
    Dim As GuiControlState state = guiState
    Dim As boolean pressed = FALSE
    
    Dim As boolean textWrap = true           ' TODO: Word-Wrap vs Char-Wrap -> textWrapMode { NO_WRAP_LOCK, NO_WRAP_OVERFLOW, CHAR_WRAP, WORD_WRAP }

    ' WARNING: First string full traversal
    Dim As Integer codepointCount = GetCodepointsCount(text0)
    
    Dim As Integer textLen = strlen(text0)     ' Text length in bytes

    ' Update control
    '--------------------------------------------------------------------
    if ((state <> GUI_STATE_DISABLED) andalso Not guiLocked) Then
  
        Dim As Vector2 mousePoint = GetMousePosition()

        if (editMode) Then
        
            state = GUI_STATE_PRESSED
            framesCounter+=1
            
            ' TODO: Cursor position logic (mouse and keys)
            
            ' Characters selection logic
            if (selectStartCp <> -1) Then
            
                if (IsKeyDown(KEY_LEFT_SHIFT) andalso IsKeyPressed(KEY_RIGHT)) Then
                
                    selectLengthCp+=1
                    if (selectLengthCp >= (codepointCount - selectStartCp)) Then selectLengthCp = codepointCount - selectStartCp
                EndIf
                
                if (IsKeyDown(KEY_LEFT_SHIFT) AndAlso IsKeyPressed(KEY_LEFT)) then
                
                    selectLengthCp-=1
                    if (selectLengthCp < 0) Then selectLengthCp = 0
                EndIf
            EndIf
            
            Dim As Integer key = GetKeyPressed()
            
            ' TODO: On key pressed, place new character in cursor position

            ' Exit edit mode logic
            if (IsKeyPressed(KEY_ENTER) orelse (Not CheckCollisionPointRec(mousePoint, bounds) andalso IsMouseButtonPressed(0))) Then pressed = TRUE
      
        else
        
            if (CheckCollisionPointRec(mousePoint, bounds)) Then
            
                state = GUI_STATE_FOCUSED
                if (IsMouseButtonPressed(0)) Then pressed = TRUE
            EndIf
        EndIf

        if (pressed) then
        
            ' Exiting edit mode, reset temp variables
            framesCounter = 0
            cursor = Rectangle(0,0,0,0)
            
            cursorCodepoint = -1
            selectStartCp = -1
            selectLengthCp = 0
        EndIf
    EndIf
    '--------------------------------------------------------------------

    ' Draw control
    '--------------------------------------------------------------------
    DrawRectangleLinesEx(bounds, GuiGetStyle(TEXTBOX, BORDER_WIDTH), Fade(GetColor(GuiGetStyle(TEXTBOX, BORDER + (state*3))), guiAlpha))

    if (state = GUI_STATE_PRESSED) Then
    DrawRectangle(bounds.x + GuiGetStyle(TEXTBOX, BORDER_WIDTH), bounds.y + GuiGetStyle(TEXTBOX, BORDER_WIDTH), bounds.width - 2*GuiGetStyle(TEXTBOX, BORDER_WIDTH), bounds.height - 2*GuiGetStyle(TEXTBOX, BORDER_WIDTH), Fade(GetColor(GuiGetStyle(TEXTBOX, BASE_COLOR_PRESSED)), guiAlpha))
    elseif (state = GUI_STATE_DISABLED) Then 
    DrawRectangle(bounds.x + GuiGetStyle(TEXTBOX, BORDER_WIDTH), bounds.y + GuiGetStyle(TEXTBOX, BORDER_WIDTH), bounds.width - 2*GuiGetStyle(TEXTBOX, BORDER_WIDTH), bounds.height - 2*GuiGetStyle(TEXTBOX, BORDER_WIDTH), Fade(GetColor(GuiGetStyle(TEXTBOX, BASE_COLOR_DISABLED)), guiAlpha))
    EndIf
    
    Dim As Font font0 = GetFontDefault()
    
    Dim As Integer textOffsetY = 0            ' Offset between lines (on line break '\n')
    Dim As Single textOffsetX = 0.0f       ' Offset X to next character to draw
    
    Dim As Single scaleFactor = GuiGetStyle(DEFAULT, TEXT_SIZE)*2/font0.baseSize     ' Character quad scaling factor
    
    Dim cp As Integer=0
    for i As integer= 0 To textLen-1
    
        ' Get next codepoint from byte string and glyph index in font
        Dim As Integer codepointByteCount = 0
        Dim As Integer codepoint = GetNextCodepoint(@text0[i], @codepointByteCount)
        Dim As Integer index = GetGlyphIndex(font0, codepoint)
        
        Dim As Rectangle rec = Rectangle( bounds.x + textOffsetX + font0.chars[index].offsetX*scaleFactor,_
                          bounds.y + textOffsetY + font0.chars[index].offsetY*scaleFactor, _
                          font0.recs[index].width*scaleFactor, font0.recs[index].height*scaleFactor )
                          
        ' Automatic line break to wrap text inside box
        if (textWrap andalso ((rec.x + rec.width) >= (bounds.x + bounds.width))) Then
        
            textOffsetY += int((font0.baseSize + font0.baseSize/2)*scaleFactor)
            textOffsetX = 0.0f
            
            ' Recalculate drawing rectangle position
            rec = Rectangle( bounds.x + textOffsetX + font0.chars[index].offsetX*scaleFactor,_
                               bounds.y + textOffsetY + font0.chars[index].offsetY*scaleFactor,_ 
                               font0.recs[index].width*scaleFactor, font0.recs[index].height*scaleFactor )
        EndIf
        
        ' Check selected codepoint
        if (editMode) Then
        
            if (IsMouseButtonPressed(MOUSE_LEFT_BUTTON) AndAlso CheckCollisionPointRec(GetMousePosition(), rec)) Then
            
                cursor = rec
                cursorCodepoint = cp
                selectStartCp = cursorCodepoint
                selectLengthCp = 0
                
                ' TODO: Place cursor at the end if pressed out of text
            EndIf
            
            ' On mouse left button down allow text selection
            if ((selectStartCp <> -1) andalso IsMouseButtonDown(MOUSE_LEFT_BUTTON)) Then
            	If CheckCollisionPointRec(GetMousePosition(), rec) Then
            
                if (cp >= selectStartCp) Then
                	selectLengthCp = cp - selectStartCp
                elseif (cp < selectStartCp) Then
                 
                    'int temp = selectStartCp
                    'selectStartCp = cp
                    'selectLengthCp = temp - selectStartCp
                EndIf
            	EndIf
            EndIf
        endif
        

        
        if (codepoint = Asc(!"\n")) Then    ' Line break character
        
            ' NOTE: Fixed line spacing of 1.5 line-height
            ' TODO: Support custom line spacing defined by user
            textOffsetY += Int((font0.baseSize + font0.baseSize/2)*scaleFactor)
            textOffsetX = 0.0f
        
        else
        
            ' Draw codepoint glyph
            if ((codepoint <> Asc(" ")) andalso (codepoint <> Asc(!"\t")) andalso ((rec.x + rec.width) < (bounds.x + bounds.width))) Then
            
                DrawTexturePro(font0.texture, font0.recs[index], rec, Vector2( 0, 0 ), 0.0f, GetColor(GuiGetStyle(DEFAULT, TEXT_COLOR_NORMAL)))
            EndIf
            
            ' TODO: On text overflow do something... move text to the left?
        EndIf
        
        ' Draw codepoints selection from selectStartCp to selectLengthCp
        ' TODO: Consider spacing when drawing selected characters background
        if (editMode andalso (selectStartCp <> -1) andalso ((cp >= selectStartCp) andalso (cp <= (selectStartCp + selectLengthCp)))) Then DrawRectangleRec(rec, MAROON)
        
        if (font0.chars[index].advanceX = 0) Then
        	textOffsetX += (font0.recs[index].width*scaleFactor + GuiGetStyle(DEFAULT, TEXT_SPACING))
        else 
        textOffsetX += (font0.chars[index].advanceX*scaleFactor + GuiGetStyle(DEFAULT, TEXT_SPACING))
        EndIf
        i += (codepointByteCount - 1)   ' Move text bytes counter to next codepoint
        cp+=1
    Next
    
    ' Draw blinking cursor
    if (editMode andalso ((framesCounter/20) Mod 2 = 0)) Then DrawRectangleRec(cursor, Fade(GetColor(GuiGetStyle(TEXTBOX, BORDER_COLOR_PRESSED)), guiAlpha))

    'GuiDrawText(text, GetTextBounds(TEXTBOX, bounds), GuiGetStyle(TEXTBOX, TEXT_ALIGNMENT), Fade(GetColor(GuiGetStyle(TEXTBOX, TEXT + (state*3))), guiAlpha))
    '--------------------------------------------------------------------

    return pressed
End function