#include once "fbgfx.bi"
'#include once "SavePng.bi"

Const xres = 512
Const yres = 512
Const midxres = xres / 2
Const midyres = yres / 2

Screenres xres, yres, 32,, FB.GFX_ALPHA_PRIMITIVES

Const Pi = Atn(1) * 4
Const TwoPi = Atn(1) * 8
Const HalfPi = Atn(1) * 2
Const RAD = Pi / 180
Const DEG = 180 / Pi
Const isq2 = 1 / Sqr(2)

Dim Shared As Uinteger texture_tmp(xres - 1, yres - 1)

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

#define Blend_Multiply(_a, _b)    ((_a * _b) shr 8)
#define Blend_Screen(_a, _b)      (255 - ((255 - _a) * (255 - _b) Shr 8))
#define Blend_Darken(_a, _b)      (iif (_a < _b, _a, _b))
#define Blend_Lighten(_a, _b)     (iif (_a > _b, _a, _b))
#define Blend_Difference(_a, _b)  (abs(_a - _b))
#define Blend_Negation(_a, _b)    (255 - abs(255 - _a - _b))
#define Blend_Exclusion(_a, _b)   (_a + _b - (_a * _b Shr 7))
#define Blend_Overlay(_a, _b)     (Iif(_a < 128, (_a * _b) Shr 7, 255 - ((255 - _a) * (255 - _b) Shr 7)))
#define Blend_HardLight(_a, _b)   (Iif(_b < 128, (_a * _b) Shr 7, 255 - ((255 - _b) * (255 - _a) Shr 7)))

Function Blend_SoftLight(a As Integer, b As Integer) As Integer
    Dim As Integer c = a * b Shr 8
    Blend_SoftLight = c + a * (255 - ((255 - a) * (255 - b) Shr 8) - c) Shr 8
End Function

Function Blend_Additive(a As Integer, b As Integer) As Integer
    Dim As Integer c
    c = a + b
    If c > 255 Then Return 255 Else Return c
End Function

Function Blend_Subtracive(a As Integer, b As Integer) As Integer
    Dim As Integer c
    c = a + b - 256
    If c < 0 Then Return 0 Else Return c
End Function

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

#define GetR(_c) ((_c Shr 16) And &hFF)
#define GetG(_c) ((_c Shr  8) And &hFF)
#define GetB(_c) ((_c       ) And &hFF)
#define GetA(_c) ((_c Shr 24) And &hFF)

#define Min(_a, _b) (iif(_a < _b, _a, _b))
#define Max(_a, _b) (iif(_a > _b, _a, _b))

Function Interpolate(y1 As Single, y2 As Single, mu As Single) As Single
    Return y1 * (1 - mu) + y2 * mu
End Function

Function randf(a As Single, b As Single) As Single
    Function = Rnd * (b - a) + a
End Function

Function randINT(first As Integer, last As Integer) As Integer
    Function = Cint(Rnd * (last - first) + first)
End Function

Function Clamp(x As Integer) As Ubyte
    If x < 0 Then x = 0
    If x > 255 Then x = 255
    Clamp = x
End Function

Function WrapX(x As Integer) As Integer
    x Mod= xres
    While x < 0
        x += xres
    Wend
   
    Return x
End Function

Function WrapY(y As Integer) As Integer
    y Mod= yres
    While y < 0
        y += yres
    Wend
   
    Return y
End Function


#define dist_Squared(_x, _y)     (_x * _x + _y * _y)


''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Perlin
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

#define FADE(_t) (_t * _t * _t * (_t * (_t * 6 - 15) + 10))
#define NLERP(_t, _a, _b) ((_a) + (_t)*((_b)-(_a)))

Dim Shared As Ubyte perm(512) = {151,160,137,91,90,15,131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,8,99,37,_
240,21,10,23,190,6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,35,11,32,57,177,33,88,237,149,56,87,174,20,125,_
136,171,168,68,175,74,165,71,134,139,48,27,166,77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,55,46,245,_
40,244,102,143,54,65,25,63,161,1,216,80,73,209,76,132,187,208,89,18,169,200,196,135,130,116,188,159,86,164,100,109,_
198,173,186,3,64,52,217,226,250,124,123,5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,189,28,42,_
223,183,170,213,119,248,152,2,44,154,163,70,221,153,101,155,167,43,172,9,129,22,39,253,19,98,108,110,79,113,224,232,_
178,185,112,104,218,246,97,228,251,34,242,193,238,210,144,12,191,179,162,241,81,51,145,235,249,14,239,107,49,192,214,_
31,181,199,106,157,184,84,204,176,115,121,50,45,127,4,150,254,138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,_
66,215,61,156,180,151,160,137,91,90,15,131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,8,99,37,240,21,10,23,_
190,6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,35,11,32,57,177,33,88,237,149,56,87,174,20,125,136,171,168,_
68,175,74,165,71,134,139,48,27,166,77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,55,46,245,40,244,102,_
143,54,65,25,63,161,1,216,80,73,209,76,132,187,208,89,18,169,200,196,135,130,116,188,159,86,164,100,109,198,173,186,_
3,64,52,217,226,250,124,123,5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,189,28,42,223,183,170,_
213,119,248,152,2,44,154,163,70,221,153,101,155,167,43,172,9,129,22,39,253,19,98,108,110,79,113,224,232,178,185,112,_
104,218,246,97,228,251,34,242,193,238,210,144,12,191,179,162,241,81,51,145,235,249,14,239,107,49,192,214,31,181,199,_
106,157,184,84,204,176,115,121,50,45,127,4,150,254,138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,66,215,61,156,180}

Dim Shared As Single ms_grad4(512)
Dim As Single kkf(256)
For i As Integer = 0 To 255
   kkf(i) = -1.0f + 2.0f * (i / 255.0f)
Next
For i As Integer = 0 To 255
   ms_grad4(i) = kkf(perm(i)) * 0.507f
Next   

Function Noise(x As Single, y As Single, px As Integer, py As Integer) As Single
   Dim As Integer ix0 = Any, iy0 = Any, ix1 = Any, iy1 = Any
   Dim As Single fx0 = Any, fy0 = Any
   Dim As Single s = Any, t = Any, nx0 = Any, nx1 = Any, n0 = Any, n1 = Any
   
   ix0 = Cint(x - 0.5f)
   iy0 = Cint(y - 0.5f)
   
    fx0 = x - ix0
   fy0 = y - iy0
   If px < 1 Then px = 1
   If py < 1 Then py = 1
   ix1 = ((ix0 + 1) Mod px) And &hff
   iy1 = ((iy0 + 1) Mod py) And &hff
   ix0 = (ix0 Mod px) And &hff
   iy0 = (iy0 Mod py) And &hff
   
   t = FADE(fy0)
   s = FADE(fx0)
   
   nx0 = ms_grad4(perm(ix0 + perm(iy0)))
   nx1 = ms_grad4(perm(ix0 + perm(iy1)))
   n0 = NLERP( t, nx0, nx1 )
   
   nx0 = ms_grad4(perm(ix1 + perm(iy0)))
   nx1 = ms_grad4(perm(ix1 + perm(iy1)))
   n1 = NLERP(t, nx0, nx1)
   
   Return NLERP(s, n0, n1)
End Function

Function Turbulence(x As Integer, y As Integer, size As Integer) As Ubyte
    ' size must be 2 ^ n
    Dim As Single value = 0.0
    Dim As Integer initialSize = size
   
    While(size >= 1)
        value += Noise(x / size, y / size, xres / size, yres / size) * size
        size /= 2.0f
    Wend
   
    Return (128.0 * value / initialSize) + 127
End Function


Function TurbulenceABS(x As Integer, y As Integer, size As Integer) As Ubyte
    ' size must be 2 ^ n
    Dim As Single value = 0.0
    Dim As Integer initialSize = size
   
    While(size >= 1)
        value += Noise(x / size, y / size, xres / size, yres / size) * size
        size /= 2.0f
    Wend
   
    Return Abs(256.0 * value / initialSize)
End Function

Function EnhancedTurbulence(x As Integer, y As Integer) As Ubyte
    Dim As Single total = 0.0
   
    Dim As Single  frequency = 1 / 256 ' USER ADJUSTABLE
    Dim As Single  persistence = 0.65  ' USER ADJUSTABLE
    Dim As Integer octaves = 8         ' USER ADJUSTABLE
    Dim As Single  amplitude = 1       ' USER ADJUSTABLE
   
    For lvc As Integer = 0 To octaves - 1
        total = total + Noise(x * frequency, y * frequency, xres * frequency, yres * frequency) * amplitude
        frequency = frequency * 2
        amplitude = amplitude * persistence
    Next
   
    Dim As Single cloudCoverage = 0   ' USER ADJUSTABLE
    Dim As Single cloudDensity = 1    ' USER ADJUSTABLE
   
    total = (total + cloudCoverage) * cloudDensity
   
    Return Clamp((128.0 * total) + 127)
End Function

Sub Generator_PerlinNoise(array() As Uinteger, size As Integer, ntype As Integer = 1)
    Dim As Integer offsetX = randINT(0, 10000)
    Dim As Integer offsetY = randINT(0, 10000)
   
    Dim As Ubyte bw
    For i As Integer = 0 To xres - 1
        For j As Integer = 0 To yres - 1
            Select Case ntype
                Case 1
                    bw = Turbulence(i + offsetX, j + offsetY, size)
                    array(i, j) = Rgb(bw, bw, bw)
                Case 2
                    bw = TurbulenceABS(i + offsetX, j + offsetY, size)
                    array(i, j) = Rgb(bw, bw, bw)
                Case 3
                    bw = EnhancedTurbulence(i + offsetX, j + offsetY)
                    array(i, j) = Rgb(bw, bw, bw)
            End Select
        Next
    Next
End Sub





''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Cells
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


Sub Generate_Cells(minitexture() As Ubyte)
    Dim As Single dx, dy, dist
    Dim As Integer bw
   
    Dim As Integer minix = Ubound(minitexture, 1)
    Dim As Integer miniy = Ubound(minitexture, 2)
   
    Dim As Integer midx = minix * 0.5f
    Dim As Integer midy = miniy * 0.5f
   
    Dim As Single scalex = 1
    Dim As Single scaley = 1
    Dim As Single maxdist
   
    If (minix > miniy) Then scalex = miniy / minix
    If (minix < miniy) Then scaley = minix / miniy
   
    maxdist = 2 / dist_Squared((midx * scalex), (midy * scaley))
    maxdist *= 255
   
    For x As Integer = 0 To minix - 1
        For y As Integer = 0 To miniy - 1
            dx = (x - midx) * scalex
            dy = (y - midy) * scaley
           
            dist = dist_Squared(dx, dy) * maxdist
            If dist > 255 Then dist = 255
            bw = 255 - dist
            minitexture(x, y) = bw
        Next
    Next
End Sub

Sub Plot_Cells(bigtexture() As Uinteger, smalltexture() As Ubyte, plotx As Integer, ploty As Integer)
    Dim As Integer minix = Ubound(smalltexture, 1)
    Dim As Integer miniy = Ubound(smalltexture, 2)
    Dim As Integer minixhalf = minix * 0.5f
    Dim As Integer miniyhalf = miniy * 0.5f
    Dim As Integer x2, y2
    Dim As Integer bw, bw1, bw2
    Dim As Integer wrapedx, wrapedy
    For x As Integer = (plotx - minixhalf) To (plotx + minixhalf) - 1
        For y As Integer = (ploty - miniyhalf) To (ploty + miniyhalf) - 1
            wrapedx = WrapX(x)
            wrapedy = WrapY(y)
            bw1 = GetR(bigtexture(wrapedx, wrapedy))
            bw2 = smalltexture(x2, y2)
            bw = Blend_Lighten(bw1, bw2)
            bigtexture(wrapedx, wrapedy) = Rgb(bw, bw, bw)
            y2 += 1
        Next
        x2 += 1
        y2 = 0
    Next
End Sub

Sub Generator_Cells(array() As Uinteger, number As Integer = 8)
    For x As Integer = 0 To xres - 1
        For y As Integer = 0 To yres - 1
            Dim As Ubyte bw = 255 - GetR(array(x, y))
            array(x, y) = Rgb(0, 0, 0)
        Next
    Next
   
    Dim As Integer elementsizex = xres * (1 / number) * 2.15f
    Dim As Integer elementsizey = yres * (1 / number) * 2.15f
    Dim As Ubyte minitexture(elementsizex, elementsizey)
    Generate_Cells(minitexture())
   
    For i As Integer = 0 To number - 1
        For j As Integer = 0 To number - 1
            Dim As Integer x = (i / number) * xres
            Dim As Integer y = (j / number) * yres
           
            x += randf(0, 1) * elementsizex
            y += randf(0, 1) * elementsizey
            x += elementsizex / 2
            y += elementsizey / 2
            If j Mod 2 = 1 Then x += elementsizex * 0.5f
            Plot_Cells(array(), minitexture(), x, y)
        Next
    Next
    For x As Integer = 0 To xres - 1
        For y As Integer = 0 To yres - 1
            Dim As Ubyte bw = 255 - GetR(array(x, y))
            array(x, y) = Rgb(bw, bw, bw)
        Next
    Next
End Sub


''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Colorize
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

Sub Effect_Colorize(array() As Uinteger, color1 As Uinteger, color2 As Uinteger)
    Dim As Ubyte r1 = GetR(color1), g1 = GetG(color1), b1 = GetB(color1)
    Dim As Ubyte r2 = GetR(color2), g2 = GetG(color2), b2 = GetB(color2)
   
    Dim As Single dist
    For x As Integer = 0 To xres - 1
        For y As Integer = 0 To yres - 1
            dist = ((GetR(array(x, y)) / 256.0f) + (GetG(array(x, y)) / 256.0f) + (GetB(array(x, y)) / 256.0f)) / 3
            array(x, y) = Rgb(Interpolate(r1, r2, dist), Interpolate(g1, g2, dist), Interpolate(b1, b2, dist))
        Next
    Next
End Sub


''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Blend
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

Sub Effect_Blend(texture1() As Uinteger, texture2() As Uinteger, mode As Integer = 1)
    Dim As Ubyte r1, g1, b1, r2, g2, b2, r, g, b
    For x As Integer = 0 To xres - 1
        For y As Integer = 0 To yres - 1
            r1 = GetR(texture1(x, y))
            g1 = GetG(texture1(x, y))
            b1 = GetB(texture1(x, y))
            r2 = GetR(texture2(x, y))
            g2 = GetG(texture2(x, y))
            b2 = GetB(texture2(x, y))
           
            If mode = 1 Then
                r = Blend_Multiply(r1, r2)
                g = Blend_Multiply(g1, g2)
                b = Blend_Multiply(b1, b2)
            End If
            If mode = 2 Then
                r = Blend_Screen(r1, r2)
                g = Blend_Screen(g1, g2)
                b = Blend_Screen(b1, b2)
            End If
            If mode = 3 Then
                r = Blend_Darken(r1, r2)
                g = Blend_Darken(g1, g2)
                b = Blend_Darken(b1, b2)
            End If
            If mode = 4 Then
                r = Blend_Lighten(r1, r2)
                g = Blend_Lighten(g1, g2)
                b = Blend_Lighten(b1, b2)
            End If
            If mode = 5 Then
                r = Blend_Difference(r1, r2)
                g = Blend_Difference(g1, g2)
                b = Blend_Difference(b1, b2)
            End If
            If mode = 6 Then
                r = Blend_Negation(r1, r2)
                g = Blend_Negation(g1, g2)
                b = Blend_Negation(b1, b2)
            End If
            If mode = 7 Then
                r = Blend_Exclusion(r1, r2)
                g = Blend_Exclusion(g1, g2)
                b = Blend_Exclusion(b1, b2)
            End If
            If mode = 8 Then
                r = Blend_Overlay(r1, r2)
                g = Blend_Overlay(g1, g2)
                b = Blend_Overlay(b1, b2)
            End If
            If mode = 9 Then
                r = Blend_HardLight(r1, r2)
                g = Blend_HardLight(g1, g2)
                b = Blend_HardLight(b1, b2)
            End If
            If mode = 10 Then
                r = Blend_SoftLight(r1, r2)
                g = Blend_SoftLight(g1, g2)
                b = Blend_SoftLight(b1, b2)
            End If
            If mode = 11 Then
                r = Blend_Additive(r1, r2)
                g = Blend_Additive(g1, g2)
                b = Blend_Additive(b1, b2)
            End If
            If mode = 12 Then
                r = Blend_Subtracive(r1, r2)
                g = Blend_Subtracive(g1, g2)
                b = Blend_Subtracive(b1, b2)
            End If
            texture1(x, y) = Rgb(r, g, b)
            texture2(x, y) = Rgb(r, g, b)
        Next
    Next
End Sub

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Emboss
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

Sub Effect_Emboss(texture1() As Uinteger, azimuth As Single, elevation As Single, width45 As Single)
    azimuth *= RAD
    elevation *= RAD
   
    Dim As Integer Nx, Ny, Nz, Lx, Ly, Lz, Nz2, NzLz, NdotL, shade
   
    Lx = Cos(azimuth) * Cos(elevation) * 255.9f
    Ly = Sin(azimuth) * Cos(elevation) * 255.9f
    Lz = Sin(elevation) * 255.9f
   
    Nz = (6 * 255) / width45
    Nz2 = Nz * Nz
    NzLz = Nz * Lz
   
    For x As Integer = 0 To xres - 1
        For y As Integer = 0 To yres - 1
            Dim As Integer a, b, c, d
            a = GetR(texture1(WrapX(x + 1), y)) + GetG(texture1(WrapX(x + 1), y)) + GetB(texture1(WrapX(x + 1), y))
            b = GetR(texture1(WrapX(x - 1), y)) + GetG(texture1(WrapX(x - 1), y)) + GetB(texture1(WrapX(x - 1), y))
            c = GetR(texture1(x, WrapY(y + 1))) + GetG(texture1(x, WrapY(y + 1))) + GetB(texture1(x, WrapY(y + 1)))
            d = GetR(texture1(x, WrapY(y - 1))) + GetG(texture1(x, WrapY(y - 1))) + GetB(texture1(x, WrapY(y - 1)))
            nx = (a - b)
            ny = (c - d)
           
            NdotL = Nx * Lx + Ny * Ly + NzLz
            If Nx = 0 And Ny = 0 Then
                shade = Lz
            Elseif NdotL < 0 Then
                shade = 0
            Else
                shade = NdotL / Sqr(Nx * Nx + Ny * Ny + Nz2)
            End If
           
            shade = Clamp(shade)
            texture_tmp(x, y) = Rgb(shade, shade, shade)
        Next
    Next
    Effect_Blend texture1(), texture_tmp(), 1
End Sub


''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Uniform Blur
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

Sub BlurHorizontal(texture1() As Uinteger, radius As Integer)
    For y As Integer = 0 To yres - 1
        Dim As Integer totalR = 0
        Dim As Integer totalG = 0
        Dim As Integer totalB = 0
       
        ' Process entire window for first pixel
        For kx As Integer = -radius To radius
            Dim As Integer kx2 = kx
            If kx2 < 0 Then kx2 += xres
            If kx2 >= xres Then kx2 -= xres
           
            totalR += GetR(texture1(kx2, y))
            totalG += GetG(texture1(kx2, y))
            totalB += GetB(texture1(kx2, y))
        Next
       
        texture_tmp(0, y) = Rgb(totalR / (radius * 2 + 1), totalG / (radius * 2 + 1), totalB / (radius * 2 + 1))
       
        ' Subsequent pixels just update window total
        For x As Integer = 1 To xres - 1
            Dim As Integer myrad1 = x - radius - 1
            Dim As Integer myrad2 = x + radius
           
            If myrad1 < 0 Then myrad1 += xres
            If myrad2 >= xres Then myrad2 -= xres
           
            ' Subtract pixel leaving window
            totalR -= GetR(texture1(myrad1, y))
            totalG -= GetG(texture1(myrad1, y))
            totalB -= GetB(texture1(myrad1, y))
           
            ' Add pixel entering window
            totalR += GetR(texture1(myrad2, y))
            totalG += GetG(texture1(myrad2, y))
            totalB += GetB(texture1(myrad2, y))
           
            texture_tmp(x, y) = Rgb(totalR / (radius * 2 + 1), totalG / (radius * 2 + 1), totalB / (radius * 2 + 1))
        Next
    Next
   
    For x As Integer = 0 To xres - 1
        For y As Integer = 0 To yres - 1
            texture1(x, y) = texture_tmp(x, y)
        Next
    Next   
End Sub

Sub BlurVertical(texture1() As Uinteger, radius As Integer)
    For x As Integer = 0 To xres - 1
        Dim As Integer totalR = 0
        Dim As Integer totalG = 0
        Dim As Integer totalB = 0
       
        ' Process entire window for first pixel
        For ky As Integer = -radius To radius
            Dim As Integer ky2 = ky
            If ky2 < 0 Then ky2 += yres
            If ky2 >= yres Then ky2 -= yres
           
            totalR += GetR(texture1(x, ky2))
            totalG += GetG(texture1(x, ky2))
            totalB += GetB(texture1(x, ky2))
        Next
       
        texture_tmp(x, 0) = Rgb(totalR / (radius * 2 + 1), totalG / (radius * 2 + 1), totalB / (radius * 2 + 1))
       
        ' Subsequent pixels just update window total
        For y As Integer = 1 To yres - 1
            ' Subtract pixel leaving window
            Dim As Integer myrad1 = y - radius - 1
            Dim As Integer myrad2 = y + radius
           
            If myrad1 < 0 Then myrad1 += yres
            If myrad2 >= yres Then myrad2 -= yres
           
            totalR -= GetR(texture1(x, myrad1))
            totalG -= GetG(texture1(x, myrad1))
            totalB -= GetB(texture1(x, myrad1))
           
            ' Add pixel entering window
            totalR += GetR(texture1(x, myrad2))
            totalG += GetG(texture1(x, myrad2))
            totalB += GetB(texture1(x, myrad2))
           
            texture_tmp(x, y) = Rgb(totalR / (radius * 2 + 1), totalG / (radius * 2 + 1), totalB / (radius * 2 + 1))
        Next
    Next
   
    For x As Integer = 0 To xres - 1
        For y As Integer = 0 To yres - 1
            texture1(x, y) = texture_tmp(x, y)
        Next
    Next   
End Sub

Sub Uniform_Blur(texture1() As Uinteger, xradius As Integer, yradius As Integer)
    If xradius > 0 Then BlurHorizontal(texture1(), xradius)
    If yradius > 0 Then BlurVertical(texture1(), yradius)
End Sub



''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Render
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

Sub Render(texture() As Uinteger, planetID As Integer)
    Dim As Uinteger Ptr buffer = Screenptr()
   
    Screenlock
   
    For x As Integer = 0 To xres - 1
        For y As Integer = 0 To yres - 1
            buffer[x + xres * y] = texture(x, y)
        Next
    Next
   
    Screenunlock
   
'    savepng "gfx/planet_" & Str(planetID) & ".png", 0, 1
End Sub



Sub Primitive_Circle(workingtexture() As Integer, midx As Integer, midy As Integer, size As Integer)
    Dim As Single dx, dy, dist
    For x As Integer = -xres To xres - 1
        For y As Integer = -yres To yres - 1
            dx = midx - x
            dy = midy - y
            dist = Sqr(dx * dx + dy * dy)
            If dist < size Then workingtexture(WrapX(x), WrapY(y)) = Rgb(255, 255, 255) Else workingtexture(WrapX(x), WrapY(y)) = Rgb(0, 0, 0)
        Next
    Next
End Sub

Sub Effect_Colorize2(array() As Uinteger, color1 As Uinteger, color2 As Uinteger)
    Dim As Ubyte r1 = GetR(color1), g1 = GetG(color1), b1 = GetB(color1)
    Dim As Ubyte r2 = GetR(color2), g2 = GetG(color2), b2 = GetB(color2)
   
    Dim As Single dist
    For x As Integer = 0 To xres - 1
        For y As Integer = 0 To yres - 1
            If GetR(array(x, y)) > 0 And GetG(array(x, y)) > 0 And GetB(array(x, y)) > 0 Then
                dist = ((GetR(array(x, y)) / 256.0f) + (GetG(array(x, y)) / 256.0f) + (GetB(array(x, y)) / 256.0f)) / 3
                array(x, y) = Rgb(Interpolate(r1, r2, dist), Interpolate(g1, g2, dist), Interpolate(b1, b2, dist))
            End If
        Next
    Next
End Sub

Sub Generator_Flat(array() As Uinteger, mycolor As Uinteger = Rgba(255, 255, 255, 255))
    For x As Integer = 0 To xres - 1
        For y As Integer = 0 To yres - 1
            array(x, y) = mycolor
        Next
    Next
End Sub

Sub Effect_Alpha(texture1() As Uinteger, texture2() As Uinteger)
    Dim As Ubyte bw
    For x As Integer = 0 To xres - 1
        For y As Integer = 0 To yres - 1
            bw = (GetR(texture2(x, y)) + GetG(texture2(x, y)) + GetB(texture2(x, y))) / 3
            texture1(x, y) = Rgba(GetR(texture1(x, y)), GetG(texture1(x, y)), GetB(texture1(x, y)), bw)
        Next
    Next
End Sub


Dim Shared As Uinteger mytexture(xres - 1, yres - 1)
Dim Shared As Uinteger mytexture2(xres - 1, yres - 1)

Randomize Timer

For i As Integer = 1 To 20
    Dim As Integer r1, g1, b1, r2, g2, b2
    Do
        r1 = randINT(0, 255) : g1 = randINT(0, 255) : b1 = randINT(0, 255)
    Loop Until r1 > 100 Or g1 > 100 Or b1 > 100
    Do
        r2 = randINT(0, 255) : g2 = randINT(0, 255) : b2 = randINT(0, 255)
    Loop Until r2 > 100 Or g2 > 100 Or b2 > 100
   
    Generator_Flat mytexture()
    Effect_Colorize2 mytexture(), Rgb(r1, g1, b1), Rgb(r2, g2, b2)
   
    Generator_PerlinNoise mytexture2(), 128, randINT(1, 3)
   
   
    Effect_Blend mytexture(), mytexture2(), 1
   
    Generator_Cells mytexture2(), randINT(4, 16)
    r1 = randINT(0, 100) : g1 = randINT(0, 100) : b1 = randINT(0, 100)
    r2 = randINT(0, 100) : g2 = randINT(0, 100) : b2 = randINT(0, 100)
    Effect_Colorize mytexture2(), Rgb(r1, g1, b1), Rgb(r2, g2, b2)
   
    Effect_Blend mytexture(), mytexture2(), 10
    Effect_Emboss mytexture(), 0, 45, 75
   
    Render mytexture(), i
Next