#Include "crt/limits.bi"
Type INoise extends object
	Declare virtual Function noise(x As Double,y As double) As double
	Declare virtual Function noise(x As Double,y As Double,z As Double) As double
	Declare virtual Function noise(x As Double,y As double,z As Double,w As Double) As double
	value As double
End Type
Function INoise.noise(x As Double,y As double) As Double
	Return value
End Function
Function INoise.noise(x As Double,y As double,z As Double) As Double
	Return value
End Function
Function INoise.noise(x As Double,y As double,z As Double,w As Double) As Double
	Return value
End Function

type ClassicNoise extends INoise
	'' Classic Perlin noise in 3D, for comparison
	Public:
	Declare Constructor()
	Declare Destructor()
	'' Classic Perlin noise, 3D version
	Declare Function noise(x as double,  y as double,  z as double) As Double
	Private:
	'' This method is a *lot* faster than using (dim as integer)Math.floor(x)
	Declare Function fastfloor(x As Double) As Integer
	Declare Function dot( g() As Integer, x as double , y as double, z as double) As Double
	Declare Function mix(a as double, b as double, t as double) As Double
	Declare Function fade(t as double) As Double

	dim As Integer grad3(12,3) = {{1,1,0},{-1,1,0},{1,-1,0},{-1,-1,0},{1,0,1},{-1,0,1},{1,0,-1},{-1,0,-1},{0,1,1},{0,-1,1},{0,1,-1},{0,-1,-1}}

	dim As Integer p(256) = {151,160,137,91,90,15,_
	131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,8,99,37,240,21,10,23,_
	190, 6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,35,11,32,57,177,33,_
	88,237,149,56,87,174,20,125,136,171,168, 68,175,74,165,71,134,139,48,27,166,_
	77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,55,46,245,40,244,_
	102,143,54, 65,25,63,161, 1,216,80,73,209,76,132,187,208, 89,18,169,200,196,_
	135,130,116,188,159,86,164,100,109,198,173,186, 3,64,52,217,226,250,124,123,_
	5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,189,28,42,_
	223,183,170,213,119,248,152, 2,44,154,163, 70,221,153,101,155,167, 43,172,9,_
	129,22,39,253, 19,98,108,110,79,113,224,232,178,185, 112,104,218,246,97,228,_
	251,34,242,193,238,210,144,12,191,179,162,241, 81,51,145,235,249,14,239,107,_
	49,192,214, 31,181,199,106,157,184, 84,204,176,115,121,50,45,127, 4,150,254,_
	138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,66,215,61,156,180}
	'' To remove the need for index wrapping, dim as double the permutation table length
	dim As integer Ptr perm = new Integer[512]
End Type

Function ClassicNoise.noise(x as double,  y as double,  z as double) As Double
	'' Find unit grid cell containing podim as integer
	dim as integer X1 = fastfloor(x)
	dim as Integer Y1 = fastfloor(y)
	dim As Integer Z1 = fastfloor(z)
	'' Get relative xyz coordinates of podim as integer within that cell
	x = x - X1
	y = y - Y1
	z = z - Z1
	'' Wrap the dim as integer cells at 255 (smaller dim as integer period can be dim as integerroduced here)
	X1 = X1 And 255
	Y1 = Y1 And 255
	Z1 = Z1 and 255
	'' Calculate a set of eight hashed gradient indices
	dim As Integer gi000 = perm[X1+perm[Y1+perm[Z1]]] Mod 12
	dim As Integer gi001 = perm[X1+perm[Y1+perm[Z1+1]]] Mod 12
	dim As Integer gi010 = perm[X1+perm[Y1+1+perm[Z1]]] Mod 12
	dim As Integer gi011 = perm[X1+perm[Y1+1+perm[Z1+1]]] Mod 12
	dim As Integer gi100 = perm[X1+1+perm[Y1+perm[Z1]]] Mod 12
	dim As Integer gi101 = perm[X1+1+perm[Y1+perm[Z1+1]]] Mod 12
	dim As Integer gi110 = perm[X1+1+perm[Y1+1+perm[Z1]]] Mod 12
	dim As Integer gi111 = perm[X1+1+perm[Y1+1+perm[Z1+1]]] Mod 12
	'' The gradients of each corner are now:
	'' g000 = grad3[gi000]
	'' g001 = grad3[gi001]
	'' g010 = grad3[gi010]
	'' g011 = grad3[gi011]
	'' g100 = grad3[gi100]
	'' g101 = grad3[gi101]
	'' g110 = grad3[gi110]
	'' g111 = grad3[gi111]
	'' Calculate noise contributions from each of the eight corners
	dim as double n000= dot(grad3(gi000,0), x, y, z)
	dim as Double n100= dot(grad3(gi100,0), x-1, y, z)
	dim As Double n010= dot(grad3(gi010,0), x, y-1, z)
	dim As Double n110= dot(grad3(gi110,0), x-1, y-1, z)
	dim As Double n001= dot(grad3(gi001,0), x, y, z-1)
	dim As Double n101= dot(grad3(gi101,0), x-1, y, z-1)
	dim As Double n011= dot(grad3(gi011,0), x, y-1, z-1)
	dim As Double n111= dot(grad3(gi111,0), x-1, y-1, z-1)
	'' Compute the fade curve value for each of x, y, z
	dim As Double u = fade(x)
	dim As Double v = fade(y)
	dim As Double w = fade(z)
	'' dim as integererpolate along x the contributions from each of the corners
	dim As Double nx00 = mix(n000, n100, u)
	dim As Double nx01 = mix(n001, n101, u)
	dim As Double nx10 = mix(n010, n110, u)
	dim As Double nx11 = mix(n011, n111, u)
	'' dim as integererpolate the four results along y
	dim As Double nxy0 = mix(nx00, nx10, v)
	dim As Double nxy1 = mix(nx01, nx11, v)
	'' dim as integererpolate the two last results along z
	dim As Double nxyz = mix(nxy0, nxy1, w)
	value = nxyz
	return value
End Function
Constructor ClassicNoise()
for i As Integer=0 To 511
	perm[i]=p(i And 255)
Next
End Constructor
Destructor ClassicNoise()
Delete [] perm
End Destructor
Function ClassicNoise.fastfloor(x As double) As Integer
	return IIf(x>0,Int(x) , Int(x-1))
End Function
Function ClassicNoise.dot(g() As Integer, x As Double, y As Double, z As Double) As Double
	return g(0)*x + g(1)*y + g(2)*z
End Function
Function ClassicNoise.mix(a As Double, b As Double, t As Double) As Double
	return (1-t)*a + t*b
End Function
Function ClassicNoise.fade(t As Double) as Double
	return t*t*t*(t*(t*6-15)+10)
End Function







Type SimplexNoise extends INoise
	Public:
	Declare Constructor()
	Declare Destructor()
	Declare Function noise(x as double,  y as Double) As Double
	Declare Function noise(x as double,  y as double,  z as double) As Double
	Declare Function noise(x as double,  y as double,  z as Double,w as Double) As Double
	'' Simplex noise in 2D, 3D and 4D
	Private:
	Declare Function fastfloor(x As Double) As Integer
	Declare Function dot2( g() As Integer, x as double , y as Double) As Double
	Declare Function dot3( g() As Integer, x as double , y as double, z as double) As Double
	Declare Function dot4( g() As Integer, x as double , y as double, z as Double, w as Double) As Double

	Dim as integer grad3(12,3) = {{1,1,0},{-1,1,0},{1,-1,0},{-1,-1,0},{1,0,1},{-1,0,1},{1,0,-1},{-1,0,-1},{0,1,1},{0,-1,1},{0,1,-1},{0,-1,-1}}
	Dim as integer grad4(32,4)= {{0,1,1,1}, {0,1,1,-1}, {0,1,-1,1}, {0,1,-1,-1},_
	{0,-1,1,1}, {0,-1,1,-1}, {0,-1,-1,1}, {0,-1,-1,-1},_
	{1,0,1,1}, {1,0,1,-1}, {1,0,-1,1}, {1,0,-1,-1},_
	{-1,0,1,1}, {-1,0,1,-1}, {-1,0,-1,1}, {-1,0,-1,-1},_
	{1,1,0,1}, {1,1,0,-1}, {1,-1,0,1}, {1,-1,0,-1},_
	{-1,1,0,1}, {-1,1,0,-1}, {-1,-1,0,1}, {-1,-1,0,-1},_
	{1,1,1,0}, {1,1,-1,0}, {1,-1,1,0}, {1,-1,-1,0},_
	{-1,1,1,0}, {-1,1,-1,0}, {-1,-1,1,0}, {-1,-1,-1,0}}
	dim as integer p(256) = {151,160,137,91,90,15,_
	131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,8,99,37,240,21,10,23,_
	190, 6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,35,11,32,57,177,33,_
	88,237,149,56,87,174,20,125,136,171,168, 68,175,74,165,71,134,139,48,27,166,_
	77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,55,46,245,40,244,_
	102,143,54, 65,25,63,161, 1,216,80,73,209,76,132,187,208, 89,18,169,200,196,_
	135,130,116,188,159,86,164,100,109,198,173,186, 3,64,52,217,226,250,124,123,_
	5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,189,28,42,_
	223,183,170,213,119,248,152, 2,44,154,163, 70,221,153,101,155,167, 43,172,9,_
	129,22,39,253, 19,98,108,110,79,113,224,232,178,185, 112,104,218,246,97,228,_
	251,34,242,193,238,210,144,12,191,179,162,241, 81,51,145,235,249,14,239,107,_
	49,192,214, 31,181,199,106,157,184, 84,204,176,115,121,50,45,127, 4,150,254,_
	138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,66,215,61,156,180}
	'' To remove the need for index wrapping, dim as double the permutation table length
	Dim as Integer ptr perm = new Integer[512]
	'' Details can be found where this table is used, in the 4D noise method.
	dim as integer simplex(64,4) = {_
	{0,1,2,3},{0,1,3,2},{0,0,0,0},{0,2,3,1},{0,0,0,0},{0,0,0,0},{0,0,0,0},{1,2,3,0},_
	{0,2,1,3},{0,0,0,0},{0,3,1,2},{0,3,2,1},{0,0,0,0},{0,0,0,0},{0,0,0,0},{1,3,2,0},_
	{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0},_
	{1,2,0,3},{0,0,0,0},{1,3,0,2},{0,0,0,0},{0,0,0,0},{0,0,0,0},{2,3,0,1},{2,3,1,0},_
	{1,0,2,3},{1,0,3,2},{0,0,0,0},{0,0,0,0},{0,0,0,0},{2,0,3,1},{0,0,0,0},{2,1,3,0},_
	{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0},_
	{2,0,1,3},{0,0,0,0},{0,0,0,0},{0,0,0,0},{3,0,1,2},{3,0,2,1},{0,0,0,0},{3,1,2,0},_
	{2,1,0,3},{0,0,0,0},{0,0,0,0},{0,0,0,0},{3,1,0,2},{0,0,0,0},{3,2,0,1},{3,2,1,0}}

End Type
Constructor SimplexNoise()
for i As Integer=0 To 511
	perm[i]=p(i And 255)
Next
End Constructor
Destructor SimplexNoise()
Delete [] perm
perm=0
End Destructor
Function SimplexNoise.fastfloor(x As double) As Integer
	return IIf(x>0,Int(x) , Int(x-1))
End Function
Function SimplexNoise.dot2(g() As Integer, x As Double, y As Double) As Double
	return g(0)*x + g(1)*y
End Function

Function SimplexNoise.dot3(g() As Integer, x As Double, y As Double, z As Double) As Double
	return g(0)*x + g(1)*y + g(2)*z
End Function
Function SimplexNoise.dot4(g() As Integer, x As Double, y As Double, z As Double, w As Double) As Double
	return g(0)*x + g(1)*y + g(2)*z + g(3)*w
End Function


'' 2D simplex noise
Function SimplexNoise.noise(xin as double,  yin as Double) As Double
	Dim as double n0, n1, n2
	'' Noise contributions from the three corners
	'' Skew the input space to determine which simplex cell we're in
	dim as double F2 = 0.5*(Sqr(3.0)-1.0)
	dim as double s = (xin+yin)*F2
	'' Hairy factor for 2D
	dim as integer i = fastfloor(xin+s)
	dim as integer j = fastfloor(yin+s)
	dim as double G2 = (3.0-Sqr(3.0))/6.0
	dim as double t = (i+j)*G2
	dim as double XX = i-t
	'' Unskew the cell origin back to (x,y) space
	dim as double YY = j-t
	dim as double x0 = xin-XX
	'' The x,y distances from the cell origin
	dim as double y0 = yin-YY
	'' For the 2D case, the simplex shape is an equilateral triangle.
	'' Determine which simplex we are in.
	dim as integer i1, j1
	'' Offsets for second (middle) corner of simplex in (i,j) coords
	if(x0>y0) Then
		i1=1: j1=0
		'' lower triangle, XY order: (0,0)->(1,0)->(1,1)
	Else
		i1=0:j1=1
	EndIf

	'' upper triangle, YX order: (0,0)->(0,1)->(1,1)
	'' A step of (1,0) in (i,j) means a step of (1-c,-c) in (x,y), and
	'' a step of (0,1) in (i,j) means a step of (-c,1-c) in (x,y), where
	'' c = (3-sqrt(3))/6
	dim as double x1 = x0 - i1 + G2
	'' Offsets for middle corner in (x,y) unskewed coords
	dim as double y1 = y0 - j1 + G2
	dim as double x2 = x0 - 1.0 + 2.0 * G2
	'' Offsets for last corner in (x,y) unskewed coords
	dim as double y2 = y0 - 1.0 + 2.0 * G2
	'' Work out the hashed gradient indices of the three simplex corners
	dim as integer ii = i And 255
	dim as integer jj = j And 255
	dim as integer gi0 = perm[ii+perm[jj]] Mod 12
	dim as integer gi1 = perm[ii+i1+perm[jj+j1]] Mod 12
	dim as integer gi2 = perm[ii+1+perm[jj+1]] Mod 12
	'' Calculate the contribution from the three corners
	dim as double t0 = 0.5 - x0*x0-y0*y0
	if(t0<0) Then
		n0 = 0.0
	Else
		t0 *= t0
		n0 = t0 * t0 * dot2(grad3(gi0,0), x0, y0)
		'' (x,y) of grad3 used for 2D gradient
	EndIf

	dim as double t1 = 0.5 - x1*x1-y1*y1
	if(t1<0) Then
		n1 = 0.0
	Else
		t1 *= t1
		n1 = t1 * t1 * dot2(grad3(gi1,0), x1, y1)
	EndIf

	dim as double t2 = 0.5 - x2*x2-y2*y2
	if(t2<0) Then
		n2 = 0.0
	else
		t2 *= t2
		n2 = t2 * t2 * dot2(grad3(gi2,0), x2, y2)
	EndIf

	'' Add contributions from each corner to get the final noise value.
	'' The result is scaled to return values in the dim as integererval [-1,1].
	value = 70.0 * (n0 + n1 + n2)
	Return value
End Function
'' 3D simplex noise
Function SimplexNoise.noise(xin as double,  yin as double,  zin as double) As Double

	dim as double n0, n1, n2, n3
	'' Noise contributions from the four corners
	'' Skew the input space to determine which simplex cell we're in
	dim as double F3 = 1.0/3.0
	dim as double s = (xin+yin+zin)*F3
	'' Very nice and simple skew factor for 3D
	dim as integer i = fastfloor(xin+s)
	dim as integer j = fastfloor(yin+s)
	dim as integer k = fastfloor(zin+s)
	Dim as double G3 = 1.0/6.0
	'' Very nice and simple unskew factor, too
	dim as double t = (i+j+k)*G3
	dim as double XX = i-t
	'' Unskew the cell origin back to (x,y,z) space
	dim as double YY = j-t
	dim as double ZZ = k-t
	dim as double x0 = xin-XX
	'' The x,y,z distances from the cell origin
	dim as double y0 = yin-YY
	dim as double z0 = zin-ZZ
	'' For the 3D case, the simplex shape is a slightly irregular tetrahedron.
	' Determine which simplex we are in.
	dim as integer i1, j1, k1
	'' Offsets for second corner of simplex in (i,j,k) coords
	dim as integer i2, j2, k2
	'' Offsets for third corner of simplex in (i,j,k) coords
	if(x0>=y0) then
		If(y0>=z0)Then
			i1=1: j1=0: k1=0: i2=1: j2=1: k2=0  '' X Y Z order
		ElseIf(x0>=z0) Then
			i1=1: j1=0: k1=0: i2=1: j2=0: k2=1 '' X Z Y order
		Else
			i1=0: j1=0: k1=1: i2=1: j2=0: k2=1 '' Z X Y order
		EndIf
	Else  '' x0<y0
		If(y0<z0) Then
			i1=0: j1=0: k1=1: i2=0: j2=1: k2=1 '' Z Y X order
		ElseIf(x0<z0) Then
			i1=0: j1=1: k1=0: i2=0: j2=1: k2=1  '' Y Z X order
		Else
			i1=0: j1=1: k1=0: i2=1: j2=1: k2=0
		EndIf'' Y X Z orde
	EndIf

	'' A step of (1,0,0) in (i,j,k) means a step of (1-c,-c,-c) in (x,y,z),
	'' a step of (0,1,0) in (i,j,k) means a step of (-c,1-c,-c) in (x,y,z), and
	'' a step of (0,0,1) in (i,j,k) means a step of (-c,-c,1-c) in (x,y,z), where
	'' c = 1/6.
	dim as double x1 = x0 - i1 + G3
	'' Offsets for second corner in (x,y,z) coords
	dim as double y1 = y0 - j1 + G3
	dim as double z1 = z0 - k1 + G3
	dim as double x2 = x0 - i2 + 2.0*G3
	'' Offsets for third corner in (x,y,z) coords
	dim as double y2 = y0 - j2 + 2.0*G3
	dim as double z2 = z0 - k2 + 2.0*G3
	dim as double x3 = x0 - 1.0 + 3.0*G3
	'' Offsets for last corner in (x,y,z) coords
	dim as double y3 = y0 - 1.0 + 3.0*G3
	dim as double z3 = z0 - 1.0 + 3.0*G3
	'' Work out the hashed gradient indices of the four simplex corners
	dim as integer ii = i And 255
	dim as integer jj = j And 255
	dim as integer kk = k and 255
	dim as integer gi0 = perm[ii+perm[jj+perm[kk]]] Mod 12
	dim as integer gi1 = perm[ii+i1+perm[jj+j1+perm[kk+k1]]] Mod 12
	dim as integer gi2 = perm[ii+i2+perm[jj+j2+perm[kk+k2]]] Mod 12
	dim as integer gi3 = perm[ii+1+perm[jj+1+perm[kk+1]]] Mod 12
	'' Calculate the contribution from the four corners
	dim as double t0 = 0.5 - x0*x0 - y0*y0 - z0*z0
	If(t0<0) Then
		n0 = 0.0
	Else
		t0 *= t0
		n0 = t0 * t0 * dot3(grad3(gi0,0), x0, y0, z0)
	EndIf
	dim as double t1 = 0.5 - x1*x1 - y1*y1 - z1*z1
	If(t1<0) Then
		n1 = 0.0
	Else
		t1 *= t1
		n1 = t1 * t1 * dot3(grad3(gi1,0), x1, y1, z1)
	EndIf
	dim as double t2 = 0.5 - x2*x2 - y2*y2 - z2*z2
	If(t2<0) Then
		n2 = 0.0
	Else
		t2 *= t2
		n2 = t2 * t2 * dot3(grad3(gi2,0), x2, y2, z2)
	EndIf
	dim as double t3 = 0.5 - x3*x3 - y3*y3 - z3*z3
	If(t3<0) Then
		n3 = 0.0
	Else
		t3 *= t3
		n3 = t3 * t3 * dot3(grad3(gi3,0), x3, y3, z3)
	EndIf
	'' Add contributions from each corner to get the final noise value.
	'' The result is scaled to stay just inside [-1,1]
	value = 32.0*(n0 + n1 + n2 + n3)
	Return value
End Function
'' 4D simplex noise
Function SimplexNoise.noise(x as double,  y as double,  z as Double,w as Double) As Double

	'' The skewing and unskewing factors are hairy again for the 4D case
	Dim as double F4 = (Sqr(5.0)-1.0)/4.0
	dim as double G4 = (5.0-Sqr(5.0))/20.0
	dim as double n0, n1, n2, n3, n4
	'' Noise contributions from the five corners
	'' Skew the (x,y,z,w) space to determine which cell of 24 simplices we're in
	dim as double s = (x + y + z + w) * F4
	'' Factor for 4D skewing
	dim as integer i = fastfloor(x + s)
	dim as integer j = fastfloor(y + s)
	dim as integer k = fastfloor(z + s)
	dim as integer l = fastfloor(w + s)
	dim as double t = (i + j + k + l) * G4
	'' Factor for 4D unskewing
	dim as double XX = i - t
	'' Unskew the cell origin back to (x,y,z,w) space
	dim as double YY = j - t
	dim as double ZZ = k - t
	dim as double WW = l - t
	dim as double x0 = x - XX
	'' The x,y,z,w distances from the cell origin
	dim as double y0 = y - YY
	dim as double z0 = z - ZZ
	dim as double w0 = w - WW
	'' For the 4D case, the simplex is a 4D shape I won't even try to describe.
	'' To find out which of the 24 possible simplices we're in, we need to
	'' determine the magnitude ordering of x0, y0, z0 and w0.
	'' The method below is a good way of finding the ordering of x,y,z,w and
	'' then find the correct traversal order for the simplex were in.
	'' First, six pair-wise comparisons are performed between each possible pair
	'' of the four coordinates, and the results are used to add up binary bits
	'' for an dim as integereger index.
	dim as integer c1 = iif(x0 > y0,32,0)
	dim as integer c2 = IIf(x0 > z0,16,0)
	dim as integer c3 = IIf(y0 > z0, 8,0)
	dim as integer c4 = IIf(x0 > w0, 4,0)
	dim as integer c5 = IIf(y0 > w0, 2,0)
	dim as integer c6 = IIf(z0 > w0, 1,0)
	dim as integer c = c1 + c2 + c3 + c4 + c5 + c6
	dim as integer i1, j1, k1, l1
	'' The dim as integereger offsets for the second simplex corner
	dim as integer i2, j2, k2, l2
	'' The dim as integereger offsets for the third simplex corner
	dim as integer i3, j3, k3, l3
	'' The dim as integereger offsets for the fourth simplex corner
	'' simplex[c] is a 4-vector with the numbers 0, 1, 2 and 3 in some order.
	'' Many values of c will never occur, since e.g. x>y>z>w makes x<z, y<w and x<w
	'' impossible. Only the 24 indices which have non-zero entries make any sense.
	'' We use a thresholding to set the coordinates in turn from the largest magnitude.
	'' The number 3 in the "simplex" array is at the position of the largest coordinate.
	i1 = IIf(simplex(c,0)>=3,1,0)
	j1 = IIf(simplex(c,1)>=3,1,0)
	k1 = IIf(simplex(c,2)>=3,1,0)
	l1 = IIf(simplex(c,3)>=3,1,0)
	'' The number 2 in the "simplex" array is at the second largest coordinate.
	i2 = IIf(simplex(c,0)>=2,1,0)
	j2 = IIf(simplex(c,1)>=2,1,0)
	k2 = IIf(simplex(c,2)>=2,1,0)
	l2 = IIf(simplex(c,3)>=2,1,0)
	'' The number 1 in the "simplex" array is at the second smallest coordinate.
	i3 = IIf(simplex(c,0)>=1,1,0)
	j3 = IIf(simplex(c,1)>=1,1,0)
	k3 = IIf(simplex(c,2)>=1,1,0)
	l3 = IIf(simplex(c,3)>=1,1,0)
	'' The fifth corner has all coordinate offsets = 1, so no need to look that up.
	dim as double x1 = x0 - i1 + G4
	'' Offsets for second corner in (x,y,z,w) coords
	dim as double y1 = y0 - j1 + G4
	dim as double z1 = z0 - k1 + G4
	dim as double w1 = w0 - l1 + G4
	dim as double x2 = x0 - i2 + 2.0*G4
	'' Offsets for third corner in (x,y,z,w) coords
	dim as double y2 = y0 - j2 + 2.0*G4
	dim as double z2 = z0 - k2 + 2.0*G4
	dim as double w2 = w0 - l2 + 2.0*G4
	dim as double x3 = x0 - i3 + 3.0*G4
	'' Offsets for fourth corner in (x,y,z,w) coords
	dim as double y3 = y0 - j3 + 3.0*G4
	dim as double z3 = z0 - k3 + 3.0*G4
	dim as double w3 = w0 - l3 + 3.0*G4
	dim as double x4 = x0 - 1.0 + 4.0*G4
	'' Offsets for last corner in (x,y,z,w) coords
	dim as double y4 = y0 - 1.0 + 4.0*G4
	dim as double z4 = z0 - 1.0 + 4.0*G4
	dim as double w4 = w0 - 1.0 + 4.0*G4
	'' Work out the hashed gradient indices of the five simplex corners
	dim as integer ii = i And 255
	dim as integer jj = j And 255
	dim as integer kk = k And 255
	dim as integer ll = l And 255
	dim as integer gi0 = perm[ii+perm[jj+perm[kk+perm[ll]]]] Mod 32
	dim as integer gi1 = perm[ii+i1+perm[jj+j1+perm[kk+k1+perm[ll+l1]]]] Mod 32
	dim as integer gi2 = perm[ii+i2+perm[jj+j2+perm[kk+k2+perm[ll+l2]]]] Mod 32
	dim as integer gi3 = perm[ii+i3+perm[jj+j3+perm[kk+k3+perm[ll+l3]]]] mod 32
	dim as integer gi4 = perm[ii+1+perm[jj+1+perm[kk+1+perm[ll+1]]]] Mod 32
	'' Calculate the contribution from the five corners
	dim as double t0 = 0.5 - x0*x0 - y0*y0 - z0*z0 - w0*w0
	If(t0<0) Then
		n0 = 0.0
	Else
		t0 *= t0
		n0 = t0 * t0 * dot4(grad4(gi0,0), x0, y0, z0, w0)
	EndIf
	dim As double t1 = 0.5 - x1*x1 - y1*y1 - z1*z1 - w1*w1
	If(t1<0) Then
		n1 = 0.0
	Else
		t1 *= t1
		n1 = t1 * t1 * dot4(grad4(gi1,0), x1, y1, z1, w1)
	EndIf
	dim as double t2 = 0.5 - x2*x2 - y2*y2 - z2*z2 - w2*w2
	If(t2<0) Then
		n2 = 0.0
	Else
		t2 *= t2
		n2 = t2 * t2 * dot4(grad4(gi2,0), x2, y2, z2, w2)
	EndIf
	dim as double t3 = 0.5 - x3*x3 - y3*y3 - z3*z3 - w3*w3
	If(t3<0) Then
		n3 = 0.0
	Else
		t3 *= t3
		n3 = t3 * t3 * dot4(grad4(gi3,0), x3, y3, z3, w3)
	EndIf
	dim as double t4 = 0.5 - x4*x4 - y4*y4 - z4*z4 - w4*w4
	If(t4<0) Then
		n4 = 0.0
	Else
		t4 *= t4
		n4 = t4 * t4 * dot4(grad4(gi4,0), x4, y4, z4, w4)
	EndIf
	'' Sum up and scale the result to cover the range [-1,1]
	value = 27.0 * (n0 + n1 + n2 + n3 + n4)
	Return value
End Function




'/*
' * OpenSimplex Noise in FreeBASIC.
' * by Kurt Spencer
' *
' * v1.1 (October 5, 2014)
' * - Added 2D and 4D implementations.
' * - Proper gradient sets for all dimensions, from a
' *   dimensionally-generalizable scheme with an actual
' *   rhyme and reason behind it.
' * - Removed default permutation array in favor of
' *   default seed.
' * - Changed seed-based constructor to be independent
' *   of any particular randomization library, so results
' *   will be the same when ported to other languages.
' */

type OpenSimplexNoise extends INoise
	Public:
	Declare Constructor()
	Declare Constructor(seed As LongInt)
	Declare Destructor()
	Declare Function noise(x As Double ,y As double) As Double
	Declare Function noise(x As Double ,y As Double, z As double) As Double
	Declare Function noise(x As Double ,y As double, z As Double, w As Double) As Double
	Private:
	Declare Function fastfloor(x As double) As Integer
	Declare Function extrapolate2(xsb As Integer, ysb As Integer, dx As Double, dy As Double) As Double
	Declare Function extrapolate3(xsb As Integer, ysb As Integer, zsb As Integer, dx As Double, dy As Double, dz As Double)As Double
	Declare Function extrapolate4(xsb As integer, ysb As Integer, zsb As Integer, wsb As Integer, dx As Double, dy As Double, dz As Double, dw As Double)As Double

	dim as double STRETCH_CONSTANT_2D = -0.211324865405187    ''(1/sqr(2+1)-1)/2
	dim As double SQUISH_CONSTANT_2D = 0.366025403784439      ''(sqr(2+1)-1)/2
	dim As double STRETCH_CONSTANT_3D = -1.0 / 6.0              ''(1/sqr(3+1)-1)/3
	dim As double SQUISH_CONSTANT_3D = 1.0 / 3.0                ''(sqr(3+1)-1)/3
	dim As double STRETCH_CONSTANT_4D = -0.138196601125011    ''(1/sqr(4+1)-1)/4
	dim As double SQUISH_CONSTANT_4D = 0.309016994374947      ''(sqr(4+1)-1)/4

	dim As double NORM_CONSTANT_2D = 47
	dim As double NORM_CONSTANT_3D = 103
	dim As double NORM_CONSTANT_4D = 30

	dim as uLongint DEFAULT_SEED = 0

	Dim As Short Ptr perm
	Dim As Short ptr permGradIndex3D

	''Gradients for 2D. They approximate the directions to the
	''vertices of an octagon from the center.
	Dim As byte gradients2D(8,2) = {_
	{5,  2},   {2,  5},_
	{-5,  2},   {-2,  5},_
	{5, -2},    {2, -5},_
	{-5, -2},   {-2, -5} _
	}

	''Gradients for 3D. They approximate the directions to the
	''vertices of a rhombicuboctahedron from the center, skewed so
	''that the triangular and square facets can be inscribed inside
	''circles of the same radius.
	Dim As byte gradients3D(24,3) = {_
	{-11,  4,  4},    { -4,  11,  4},   { -4,  4,  11},_
	{11,  4,  4},     { 4,  11,  4},    { 4,  4,  11},_
	{-11, -4,  4},    { -4, -11,  4},   { -4, -4,  11},_
	{11, -4,  4},      {4, -11,  4},     {4, -4,  11},_
	{-11,  4, -4},    { -4,  11, -4},   {-4,  4, -11},_
	{11,  4, -4},      {4,  11, -4},    { 4,  4, -11},_
	{-11, -4, -4},    {-4, -11, -4},    {-4, -4, -11},_
	{11, -4, -4},      {4, -11, -4},    { 4, -4, -11} _
	}

	''Gradients for 4D. They approximate the directions to the
	''vertices of a disprismatotesseractihexadecachoron from the center,
	''skewed so that the tetrahedral and cubic facets can be inscribed inside
	''spheres of the same radius.
	Dim As Byte gradients4D(64,4) = {_
	{3,  1,  1,  1},     { 1,  3,  1,  1},     { 1,  1,  3,  1},     { 1,  1,  1,  3},_
	{-3,  1,  1,  1},    { -1,  3,  1,  1},    { -1,  1,  3,  1},    { -1,  1,  1,  3},_
	{3, -1,  1,  1},     { 1, -3,  1,  1},     { 1, -1,  3,  1},     { 1, -1,  1,  3},_
	{-3, -1,  1,  1},    { -1, -3,  1,  1},    { -1, -1,  3,  1},    { -1, -1,  1,  3},_
	{3,  1, -1,  1},     { 1,  3, -1,  1},     { 1,  1, -3,  1},     { 1,  1, -1,  3},_
	{-3,  1, -1,  1},    { -1,  3, -1,  1},    { -1,  1, -3,  1},    { -1,  1, -1,  3},_
	{3, -1, -1,  1},     { 1, -3, -1,  1},     { 1, -1, -3,  1},     { 1, -1, -1,  3},_
	{-3, -1, -1,  1},    { -1, -3, -1,  1},    { -1, -1, -3,  1},    { -1, -1, -1,  3},_
	{3,  1,  1, -1},      {1,  3,  1, -1},     { 1,  1,  3, -1},     { 1,  1,  1, -3},_
	{-3,  1,  1, -1},     {-1,  3,  1, -1},    { -1,  1,  3, -1},    { -1,  1,  1, -3},_
	{3, -1,  1, -1},      {1, -3,  1, -1},     { 1, -1,  3, -1},     { 1, -1,  1, -3},_
	{-3, -1,  1, -1},     {-1, -3,  1, -1},    { -1, -1,  3, -1},    { -1, -1,  1, -3},_
	{3,  1, -1, -1},      {1,  3, -1, -1},     { 1,  1, -3, -1},     { 1,  1, -1, -3},_
	{-3,  1, -1, -1},    { -1,  3, -1, -1},    { -1,  1, -3, -1},    { -1,  1, -1, -3},_
	{3, -1, -1, -1},     { 1, -3, -1, -1},     { 1, -1, -3, -1},     { 1, -1, -1, -3}, _
	{-3, -1, -1, -1},     {-1, -3, -1, -1},    { -1, -1, -3, -1},    { -1, -1, -1, -3} _
	}
End type


Constructor OpenSimplexNoise()
Dim seed As ULongInt=DEFAULT_SEED
perm = new Short[256]
permGradIndex3D = new Short[256]
Dim As Short Ptr source = new Short[256]
for i As Integer=0 To 255
	source[i] = i
next
seed = seed * 6364136223846793005ull + 1442695040888963407ull
seed = seed * 6364136223846793005ull + 1442695040888963407ull
seed = seed * 6364136223846793005ull + 1442695040888963407ull
for i As Integer=255 To 0 Step -1
	seed = seed * 6364136223846793005ull + 1442695040888963407ull
	Dim As integer r = Int((seed + 31) Mod (i + 1))
	if (r < 0) Then
		r += (i + 1)
	EndIf
	perm[i] = source[r]
	permGradIndex3D[i] = Cast(Short,((perm[i] Mod (24)) ))
	source[r] = source[i]
Next
Delete [] source
source=0
End Constructor

''Initializes the class using a permutation array generated from a 64-bit seed.
''Generates a proper permutation (i.e. doesn't merely perform N successive pair swaps on a base array)
''Uses a simple 64-bit LCG.
Constructor OpenSimplexNoise(seed As LongInt)
perm = new short[256]
permGradIndex3D = new short[256]
Dim As Short Ptr source = new short[256]
for i As UInteger=0 To 255
	source[i] = i
next
seed = seed * 6364136223846793005ull + 1442695040888963407ull
seed = seed * 6364136223846793005ull + 1442695040888963407ull
seed = seed * 6364136223846793005ull + 1442695040888963407ull
for i As Integer=255 To 0 Step -1
	seed = seed * 6364136223846793005ull + 1442695040888963407ull
	Dim As integer r = Int((seed + 31) Mod (i + 1))
	if (r < 0) Then
		r += (i + 1)
	EndIf
	perm[i] = source[r]
	permGradIndex3D[i] = cast(short,((perm[i] Mod (24)) ))
	source[r] = source[i]
Next
Delete [] source
source=0
End Constructor
Destructor OpenSimplexNoise()
Delete [] perm
perm=0
Delete [] permGradIndex3D
permGradIndex3D=0
End Destructor
Function OpenSimplexNoise.fastfloor(x As double) As Integer
	return IIf(x>=0,Int(x) , Int(x-1))
End Function
Function OpenSimplexNoise.extrapolate2(xsb As Integer, ysb As Integer, dx As Double, dy As Double) As Double
	Dim As integer index = perm[(perm[xsb and &hFF] + ysb) and &hFF] Mod 8
	return gradients2D(index,0) * dx + gradients2D(index,1) * dy
End Function

Function OpenSimplexNoise.extrapolate3(xsb As Integer, ysb As Integer, zsb As Integer, dx As Double, dy As Double, dz As Double)As Double
	Dim As integer index = permGradIndex3D[(perm[(perm[xsb And &hFF] + ysb) And &hFF] + zsb) And &hFF] Mod 24
	return gradients3D(index,0) * dx + gradients3D(index,1) * dy + gradients3D(index, 2) * dz
End Function

Function OpenSimplexNoise.extrapolate4(xsb As integer, ysb As Integer, zsb As Integer, wsb As Integer, dx As Double, dy As Double, dz As Double, dw As Double)As Double
	Dim As integer index = perm[(perm[(perm[(perm[xsb And &hFF] + ysb) And &hFF] + zsb) and &hFF] + wsb) And &hFF] Mod 64
	return gradients4D(index,0) * dx + gradients4D(index , 1) * dy + gradients4D(index , 2) * dz + gradients4D(index , 3) * dw
End function

''2D OpenSimplex Noise.
function  OpenSimplexNoise.noise(x As Double ,y As double) As Double

	''Place input coordinates onto grid.
	Dim As Double stretchOffset = (x + y) * STRETCH_CONSTANT_2D
	Dim As Double xs = x + stretchOffset
	Dim As Double ys = y + stretchOffset

	''Floor to get grid coordinates of rhombus (stretched square) super-cell origin.
	Dim As Integer xsb = fastFloor(xs)
	Dim As Integer ysb = fastFloor(ys)

	''Skew out to get actual coordinates of rhombus origin. We'll need these later.
	Dim As Double squishOffset = (xsb + ysb) * SQUISH_CONSTANT_2D
	Dim As Double xb = xsb + squishOffset
	Dim As Double yb = ysb + squishOffset

	''Compute grid coordinates relative to rhombus origin.
	Dim As Double xins = xs - xsb
	Dim As Double yins = ys - ysb

	''Sum those together to get a value that determines which region we're in.
	Dim As Double inSum = xins + yins

	''Positions relative to origin point.
	Dim As Double dx0 = x - xb
	Dim As Double dy0 = y - yb

	''We'll be defining these inside the next block and using them afterwards.
	Dim As Double dx_ext, dy_ext
	Dim As Integer xsv_ext, ysv_ext

	Dim As Double value = 0

	''Contribution (1,0)
	Dim As Double dx1 = dx0 - 1.0 - SQUISH_CONSTANT_2D
	Dim As Double dy1 = dy0 - 0.0 - SQUISH_CONSTANT_2D
	Dim As Double attn1 = 2.0 - (dx1 * dx1) - (dy1 * dy1)
	if (attn1 > 0.0) Then
		attn1 *= attn1
		value += (attn1 * attn1) * extrapolate2(xsb + 1, ysb + 0, dx1, dy1)
	EndIf

	''Contribution (0,1)
	Dim As Double dx2 = dx0 - 0.0 - SQUISH_CONSTANT_2D
	Dim As Double dy2 = dy0 - 1.0 - SQUISH_CONSTANT_2D
	Dim As Double attn2 = 2.0 - dx2 * dx2 - dy2 * dy2
	if (attn2 > 0) Then
		attn2 *= attn2
		value += (attn2 * attn2) * extrapolate2(xsb + 0, ysb + 1, dx2, dy2)
	endif

	if (inSum <= 1.0) then ''We're inside the triangle (2-Simplex) at (0,0)
		Dim As Double zins = 1.0 - inSum
		if ((zins > xins) or (zins > yins)) then ''(0,0) is one of the closest two triangular vertices
			if (xins > yins) Then
				xsv_ext = xsb + 1.0
				ysv_ext = ysb - 1.0
				dx_ext = dx0 - 1.0
				dy_ext = dy0 + 1.0
			else
				xsv_ext = xsb - 1.0
				ysv_ext = ysb + 1.0
				dx_ext = dx0 + 1.0
				dy_ext = dy0 - 1.0
			EndIf
		Else ''(1,0) and (0,1) are the closest two vertices.
			xsv_ext = xsb + 1.0
			ysv_ext = ysb + 1.0
			dx_ext = dx0 - 1.0 - 2.0 * SQUISH_CONSTANT_2D
			dy_ext = dy0 - 1.0 - 2.0 * SQUISH_CONSTANT_2D
		EndIf
	else ''We're inside the triangle (2-Simplex) at (1,1)
		Dim As Double zins = 2.0 - inSum
		if ((zins < xins) or (zins < yins)) Then ''(0,0) is one of the closest two triangular vertices
			if (xins > yins) Then
				xsv_ext = xsb + 2.0
				ysv_ext = ysb + 0.0
				dx_ext = dx0 - 2.0 - 2.0 * SQUISH_CONSTANT_2D
				dy_ext = dy0 + 0.0 - 2.0 * SQUISH_CONSTANT_2D
			Else
				xsv_ext = xsb + 0.0
				ysv_ext = ysb + 2.0
				dx_ext = dx0 + 0.0 - 2.0 * SQUISH_CONSTANT_2D
				dy_ext = dy0 - 2.0 - 2.0 * SQUISH_CONSTANT_2D
			endif
		else  ''(1,0) and (0,1) are the closest two vertices.
			dx_ext = dx0
			dy_ext = dy0
			xsv_ext = xsb
			ysv_ext = ysb
		EndIf
		xsb += 1.0
		ysb += 1.0
		dx0 = dx0 - 1.0 - 2.0 * SQUISH_CONSTANT_2D
		dy0 = dy0 - 1.0 - 2.0 * SQUISH_CONSTANT_2D
	EndIf

	''Contribution (0,0) or (1,1)
	Dim As Double attn0 = 2.0 - dx0 * dx0 - dy0 * dy0
	if (attn0 > 0.0) Then
		attn0 *= attn0
		value += (attn0 * attn0) * extrapolate2(xsb, ysb, dx0, dy0)
	EndIf

	''Extra Vertex
	Dim As Double attn_ext = 2.0 - dx_ext * dx_ext - dy_ext * dy_ext
	if (attn_ext > 0.0)Then
		attn_ext *= attn_ext
		value += (attn_ext * attn_ext) * extrapolate2(xsv_ext, ysv_ext, dx_ext, dy_ext)
	endif
	value/=NORM_CONSTANT_2D
	return value
End Function

''3D OpenSimplex Noise.
Function OpenSimplexNoise.noise(x As Double ,y As Double, z As double) As Double

	''Place input coordinates on simplectic honeycomb.
	Dim As Double stretchOffset = (x + y + z) * STRETCH_CONSTANT_3D
	Dim As Double xs = x + stretchOffset
	Dim As Double ys = y + stretchOffset
	Dim As Double zs = z + stretchOffset

	''Floor to get simplectic honeycomb coordinates of rhombohedron (stretched cube) super-cell origin.
	Dim As integer xsb = fastFloor(xs)
	Dim As Integer ysb = fastFloor(ys)
	Dim As Integer zsb = fastFloor(zs)

	''Skew out to get actual coordinates of rhombohedron origin. We'll need these later.
	Dim As Double squishOffset = (xsb + ysb + zsb) * SQUISH_CONSTANT_3D
	Dim As Double xb = xsb + squishOffset
	Dim As Double yb = ysb + squishOffset
	Dim As Double zb = zsb + squishOffset

	''Compute simplectic honeycomb coordinates relative to rhombohedral origin.
	Dim As Double xins = xs - xsb
	Dim As Double yins = ys - ysb
	Dim As Double zins = zs - zsb

	''Sum those together to get a value that determines which region we're in.
	Dim As Double inSum = xins + yins + zins

	''Positions relative to origin point.
	Dim As Double dx0 = x - xb
	Dim As Double dy0 = y - yb
	Dim As Double dz0 = z - zb

	''We'll be defining these inside the next block and using them afterwards.
	Dim As Double dx_ext0, dy_ext0, dz_ext0
	Dim As Double dx_ext1, dy_ext1, dz_ext1
	Dim As Integer xsv_ext0, ysv_ext0, zsv_ext0
	Dim As Integer xsv_ext1, ysv_ext1, zsv_ext1

	Dim As Double value = 0
	if (inSum <= 1) Then ''We're inside the tetrahedron (3-Simplex) at (0,0,0)

		''Determine which two of (0,0,1), (0,1,0), (1,0,0) are closest.
		Dim As Byte aPoint = &h01
		Dim As Double aScore = xins
		Dim As Byte bPoint = &h02
		Dim As Double bScore = yins
		if (aScore >= bScore And zins > bScore) Then
			bScore = zins
			bPoint = &h04
		elseif (aScore < bScore and zins > aScore) Then
			aScore = zins
			aPoint = &h04
		endif

		''Now we determine the two lattice points not part of the tetrahedron that may contribute.
		''This depends on the closest two tetrahedral vertices, including (0,0,0)
		Dim As Double wins = 1 - inSum
		if (wins > aScore Or wins > bScore) Then ''(0,0,0) is one of the closest two tetrahedral vertices.
			Dim As Byte c = iif(bScore > aScore , bPoint , aPoint) ''Our other closest vertex is the closest out of a and b.

			if ((c and &h01) = 0) Then
				xsv_ext0 = xsb - 1
				xsv_ext1 = xsb
				dx_ext0 = dx0 + 1
				dx_ext1 = dx0
			Else
				xsv_ext0 = xsv_ext1 = xsb + 1
				dx_ext0 = dx_ext1 = dx0 - 1
			endif

			if ((c and &h02) = 0) Then
				ysv_ext0 = ysv_ext1 = ysb
				dy_ext0 = dy_ext1 = dy0
				if ((c and &h01) = 0) Then
					ysv_ext1 -= 1
					dy_ext1 += 1
				Else
					ysv_ext0 -= 1
					dy_ext0 += 1
				endif
			Else
				ysv_ext0 = ysv_ext1 = ysb + 1
				dy_ext0 = dy_ext1 = dy0 - 1
			EndIf

			if ((c and &h04) = 0) then
				zsv_ext0 = zsb
				zsv_ext1 = zsb - 1
				dz_ext0 = dz0
				dz_ext1 = dz0 + 1
			Else
				zsv_ext0 = zsv_ext1 = zsb + 1
				dz_ext0 = dz_ext1 = dz0 - 1
			EndIf
		Else ''(0,0,0) is not one of the closest two tetrahedral vertices.
			Dim As Byte c = aPoint And bPoint'cast(Byte,(aPoint or bPoint)) ''Our two extra vertices are determined by the closest two.

			if ((c and &h01) = 0) Then
				xsv_ext0 = xsb
				xsv_ext1 = xsb - 1
				dx_ext0 = dx0 - 2 * SQUISH_CONSTANT_3D
				dx_ext1 = dx0 + 1 - SQUISH_CONSTANT_3D
			Else
				xsv_ext0 = xsv_ext1 = xsb + 1
				dx_ext0 = dx0 - 1 - 2 * SQUISH_CONSTANT_3D
				dx_ext1 = dx0 - 1 - SQUISH_CONSTANT_3D
			EndIf

			if ((c and &h02) = 0) Then
				ysv_ext0 = ysb
				ysv_ext1 = ysb - 1
				dy_ext0 = dy0 - 2 * SQUISH_CONSTANT_3D
				dy_ext1 = dy0 + 1 - SQUISH_CONSTANT_3D
			Else
				ysv_ext0 = ysv_ext1 = ysb + 1
				dy_ext0 = dy0 - 1 - 2 * SQUISH_CONSTANT_3D
				dy_ext1 = dy0 - 1 - SQUISH_CONSTANT_3D
			EndIf

			if ((c and &h04) = 0) Then
				zsv_ext0 = zsb
				zsv_ext1 = zsb - 1
				dz_ext0 = dz0 - 2 * SQUISH_CONSTANT_3D
				dz_ext1 = dz0 + 1 - SQUISH_CONSTANT_3D
			Else
				zsv_ext0 = zsv_ext1 = zsb + 1
				dz_ext0 = dz0 - 1 - 2 * SQUISH_CONSTANT_3D
				dz_ext1 = dz0 - 1 - SQUISH_CONSTANT_3D
			endif
		endif

		''Contribution (0,0,0)
		Dim As Double attn0 = 2 - dx0 * dx0 - dy0 * dy0 - dz0 * dz0
		if (attn0 > 0) Then
			attn0 *= attn0
			value += attn0 * attn0 * extrapolate3(xsb + 0, ysb + 0, zsb + 0, dx0, dy0, dz0)
		endif

		''Contribution (1,0,0)
		Dim As Double dx1 = dx0 - 1 - SQUISH_CONSTANT_3D
		Dim As Double dy1 = dy0 - 0 - SQUISH_CONSTANT_3D
		Dim As Double dz1 = dz0 - 0 - SQUISH_CONSTANT_3D
		Dim As Double attn1 = 2 - dx1 * dx1 - dy1 * dy1 - dz1 * dz1
		if (attn1 > 0) Then
			attn1 *= attn1
			value += attn1 * attn1 * extrapolate3(xsb + 1, ysb + 0, zsb + 0, dx1, dy1, dz1)
		EndIf

		''Contribution (0,1,0)
		Dim As Double dx2 = dx0 - 0 - SQUISH_CONSTANT_3D
		Dim As Double dy2 = dy0 - 1 - SQUISH_CONSTANT_3D
		Dim As Double dz2 = dz1
		Dim As Double attn2 = 2 - dx2 * dx2 - dy2 * dy2 - dz2 * dz2
		if (attn2 > 0) Then
			attn2 *= attn2
			value += attn2 * attn2 * extrapolate3(xsb + 0, ysb + 1, zsb + 0, dx2, dy2, dz2)
		EndIf

		''Contribution (0,0,1)
		Dim As Double dx3 = dx2
		Dim As Double dy3 = dy1
		Dim As Double dz3 = dz0 - 1 - SQUISH_CONSTANT_3D
		Dim As Double attn3 = 2 - dx3 * dx3 - dy3 * dy3 - dz3 * dz3
		if (attn3 > 0) Then
			attn3 *= attn3
			value += attn3 * attn3 * extrapolate3(xsb + 0, ysb + 0, zsb + 1, dx3, dy3, dz3)
		EndIf
	ElseIf (inSum >= 2) Then ''We're inside the tetrahedron (3-Simplex) at (1,1,1)

		''Determine which two tetrahedral vertices are the closest, out of (1,1,0), (1,0,1), (0,1,1) but not (1,1,1).
		Dim As Byte aPoint = &h06
		Dim As Double aScore = xins
		Dim As Byte bPoint = &h05
		Dim As Double bScore = yins
		if (aScore <= bScore And zins < bScore) Then
			bScore = zins
			bPoint = &h03
		elseif (aScore > bScore And zins < aScore) Then
			aScore = zins
			aPoint = &h03
		endif

		''Now we determine the two lattice points not part of the tetrahedron that may contribute.
		''This depends on the closest two tetrahedral vertices, including (1,1,1)
		Dim As Double wins = 3 - inSum
		if (wins < aScore Or wins < bScore) Then ''(1,1,1) is one of the closest two tetrahedral vertices.
			Dim As uByte c = iif(bScore < aScore , bPoint , aPoint) ''Our other closest vertex is the closest out of a and b.

			if ((c and &h01) <> 0) then
				xsv_ext0 = xsb + 2
				xsv_ext1 = xsb + 1
				dx_ext0 = dx0 - 2 - 3 * SQUISH_CONSTANT_3D
				dx_ext1 = dx0 - 1 - 3 * SQUISH_CONSTANT_3D
			Else
				xsv_ext0 = xsv_ext1 = xsb
				dx_ext0 = dx_ext1 = dx0 - 3 * SQUISH_CONSTANT_3D
			EndIf

			if ((c and &h02) <> 0) Then
				ysv_ext0 = ysv_ext1 = ysb + 1
				dy_ext0 = dy_ext1 = dy0 - 1 - 3 * SQUISH_CONSTANT_3D
				if ((c and &h01) <> 0) Then
					ysv_ext1 += 1
					dy_ext1 -= 1
				Else
					ysv_ext0 += 1
					dy_ext0 -= 1
				EndIf
			Else
				ysv_ext0 = ysv_ext1 = ysb
				dy_ext0 = dy_ext1 = dy0 - 3 * SQUISH_CONSTANT_3D
			endif

			if ((c and &h04) <> 0) Then
				zsv_ext0 = zsb + 1
				zsv_ext1 = zsb + 2
				dz_ext0 = dz0 - 1 - 3 * SQUISH_CONSTANT_3D
				dz_ext1 = dz0 - 2 - 3 * SQUISH_CONSTANT_3D
			Else
				zsv_ext0 = zsv_ext1 = zsb
				dz_ext0 = dz_ext1 = dz0 - 3 * SQUISH_CONSTANT_3D
			EndIf
		Else ''(1,1,1) is not one of the closest two tetrahedral vertices.
			Dim As Byte c = aPoint And bPoint'cast(byte,(aPoint And bPoint)) ''Our two extra vertices are determined by the closest two.

			if ((c and &h01) <> 0) Then
				xsv_ext0 = xsb + 1
				xsv_ext1 = xsb + 2
				dx_ext0 = dx0 - 1 - SQUISH_CONSTANT_3D
				dx_ext1 = dx0 - 2 - 2 * SQUISH_CONSTANT_3D
			Else
				xsv_ext0 = xsv_ext1 = xsb
				dx_ext0 = dx0 - SQUISH_CONSTANT_3D
				dx_ext1 = dx0 - 2 * SQUISH_CONSTANT_3D
			EndIf

			if ((c and &h02) <> 0) Then
				ysv_ext0 = ysb + 1
				ysv_ext1 = ysb + 2
				dy_ext0 = dy0 - 1 - SQUISH_CONSTANT_3D
				dy_ext1 = dy0 - 2 - 2 * SQUISH_CONSTANT_3D
			Else
				ysv_ext0 = ysv_ext1 = ysb
				dy_ext0 = dy0 - SQUISH_CONSTANT_3D
				dy_ext1 = dy0 - 2 * SQUISH_CONSTANT_3D
			EndIf

			if ((c and &h04) <> 0) Then
				zsv_ext0 = zsb + 1
				zsv_ext1 = zsb + 2
				dz_ext0 = dz0 - 1 - SQUISH_CONSTANT_3D
				dz_ext1 = dz0 - 2 - 2 * SQUISH_CONSTANT_3D
			Else
				zsv_ext0 = zsv_ext1 = zsb
				dz_ext0 = dz0 - SQUISH_CONSTANT_3D
				dz_ext1 = dz0 - 2 * SQUISH_CONSTANT_3D
			EndIf
		endif

		''Contribution (1,1,0)
		Dim As Double dx3 = dx0 - 1 - 2 * SQUISH_CONSTANT_3D
		Dim As Double dy3 = dy0 - 1 - 2 * SQUISH_CONSTANT_3D
		Dim As Double dz3 = dz0 - 0 - 2 * SQUISH_CONSTANT_3D
		Dim As Double attn3 = 2 - dx3 * dx3 - dy3 * dy3 - dz3 * dz3
		if (attn3 > 0) Then
			attn3 *= attn3
			value += attn3 * attn3 * extrapolate3(xsb + 1, ysb + 1, zsb + 0, dx3, dy3, dz3)
		EndIf

		''Contribution (1,0,1)
		Dim As Double dx2 = dx3
		Dim As Double dy2 = dy0 - 0 - 2 * SQUISH_CONSTANT_3D
		Dim As Double dz2 = dz0 - 1 - 2 * SQUISH_CONSTANT_3D
		Dim As Double attn2 = 2 - dx2 * dx2 - dy2 * dy2 - dz2 * dz2
		if (attn2 > 0) Then
			attn2 *= attn2
			value += attn2 * attn2 * extrapolate3(xsb + 1, ysb + 0, zsb + 1, dx2, dy2, dz2)
		EndIf

		''Contribution (0,1,1)
		Dim As Double dx1 = dx0 - 0 - 2 * SQUISH_CONSTANT_3D
		Dim As Double dy1 = dy3
		Dim As Double dz1 = dz2
		Dim As Double attn1 = 2 - dx1 * dx1 - dy1 * dy1 - dz1 * dz1
		if (attn1 > 0) Then
			attn1 *= attn1
			value += attn1 * attn1 * extrapolate3(xsb + 0, ysb + 1, zsb + 1, dx1, dy1, dz1)
		endif

		''Contribution (1,1,1)
		dx0 = dx0 - 1 - 3 * SQUISH_CONSTANT_3D
		dy0 = dy0 - 1 - 3 * SQUISH_CONSTANT_3D
		dz0 = dz0 - 1 - 3 * SQUISH_CONSTANT_3D
		Dim As Double attn0 = 2 - dx0 * dx0 - dy0 * dy0 - dz0 * dz0
		if (attn0 > 0) Then
			attn0 *= attn0
			value += attn0 * attn0 * extrapolate3(xsb + 1, ysb + 1, zsb + 1, dx0, dy0, dz0)
		EndIf
	Else  ''We're inside the octahedron (Rectified 3-Simplex) in between.
		Dim As Double aScore
		Dim As Byte aPoint
		Dim As boolean aIsFurtherSide=FALSE
		Dim As Double bScore
		Dim As Byte bPoint
		Dim As boolean bIsFurtherSide=FALSE

		''Decide between point (0,0,1) and (1,1,0) as closest
		Dim As Double p1 = xins + yins
		if (p1 > 1) Then
			aScore = p1 - 1
			aPoint = &h03
			aIsFurtherSide = TRUE
		Else
			aScore = 1 - p1
			aPoint = &h04
			aIsFurtherSide = FALSE
		EndIf

		''Decide between point (0,1,0) and (1,0,1) as closest
		Dim As Double p2 = xins + zins
		if (p2 > 1) Then
			bScore = p2 - 1
			bPoint = &h05
			bIsFurtherSide = TRUE
		else
			bScore = 1 - p2
			bPoint = &h02
			bIsFurtherSide = FALSE
		EndIf

		''The closest out of the two (1,0,0) and (0,1,1) will replace the furthest out of the two decided above, if closer.
		Dim As Double p3 = yins + zins
		if (p3 > 1) Then
			Dim As Double score = p3 - 1
			if (aScore <= bScore And aScore < score) Then
				aScore = score
				aPoint = &h06
				aIsFurtherSide = TRUE
			elseif (aScore > bScore And bScore < score) Then
				bScore = score
				bPoint = &h06
				bIsFurtherSide = TRUE
			EndIf
		Else
			Dim As Double score = 1 - p3
			if (aScore <= bScore And aScore < score) Then
				aScore = score
				aPoint = &h01
				aIsFurtherSide = FALSE
			ElseIf (aScore > bScore And bScore < score) Then
				bScore = score
				bPoint = &h01
				bIsFurtherSide = FALSE
			EndIf
		endif

		''Where each of the two closest points are determines how the extra two vertices are calculated.
		if (aIsFurtherSide = bIsFurtherSide) Then
			if (aIsFurtherSide) Then ''Both closest points on (1,1,1) side

				''One of the two extra points is (1,1,1)
				dx_ext0 = dx0 - 1 - 3 * SQUISH_CONSTANT_3D
				dy_ext0 = dy0 - 1 - 3 * SQUISH_CONSTANT_3D
				dz_ext0 = dz0 - 1 - 3 * SQUISH_CONSTANT_3D
				xsv_ext0 = xsb + 1
				ysv_ext0 = ysb + 1
				zsv_ext0 = zsb + 1

				''Other extra point is based on the shared axis.
				Dim As Byte c = aPoint And bPoint'cast(byte,(aPoint And bPoint))
				if ((c and &h01) <> 0) Then
					dx_ext1 = dx0 - 2 - 2 * SQUISH_CONSTANT_3D
					dy_ext1 = dy0 - 2 * SQUISH_CONSTANT_3D
					dz_ext1 = dz0 - 2 * SQUISH_CONSTANT_3D
					xsv_ext1 = xsb + 2
					ysv_ext1 = ysb
					zsv_ext1 = zsb
				elseif ((c and &h02) <> 0) Then
					dx_ext1 = dx0 - 2 * SQUISH_CONSTANT_3D
					dy_ext1 = dy0 - 2 - 2 * SQUISH_CONSTANT_3D
					dz_ext1 = dz0 - 2 * SQUISH_CONSTANT_3D
					xsv_ext1 = xsb
					ysv_ext1 = ysb + 2
					zsv_ext1 = zsb
				Else
					dx_ext1 = dx0 - 2 * SQUISH_CONSTANT_3D
					dy_ext1 = dy0 - 2 * SQUISH_CONSTANT_3D
					dz_ext1 = dz0 - 2 - 2 * SQUISH_CONSTANT_3D
					xsv_ext1 = xsb
					ysv_ext1 = ysb
					zsv_ext1 = zsb + 2
				EndIf
			Else''Both closest points on (0,0,0) side

				''One of the two extra points is (0,0,0)
				dx_ext0 = dx0
				dy_ext0 = dy0
				dz_ext0 = dz0
				xsv_ext0 = xsb
				ysv_ext0 = ysb
				zsv_ext0 = zsb

				''Other extra point is based on the omitted axis.
				Dim As Byte c = aPoint And bPoint'cast(byte,(aPoint or bPoint))
				if ((c and &h01) = 0) Then
					dx_ext1 = dx0 + 1 - SQUISH_CONSTANT_3D
					dy_ext1 = dy0 - 1 - SQUISH_CONSTANT_3D
					dz_ext1 = dz0 - 1 - SQUISH_CONSTANT_3D
					xsv_ext1 = xsb - 1
					ysv_ext1 = ysb + 1
					zsv_ext1 = zsb + 1
				ElseIf ((c and &h02) = 0) then
					dx_ext1 = dx0 - 1 - SQUISH_CONSTANT_3D
					dy_ext1 = dy0 + 1 - SQUISH_CONSTANT_3D
					dz_ext1 = dz0 - 1 - SQUISH_CONSTANT_3D
					xsv_ext1 = xsb + 1
					ysv_ext1 = ysb - 1
					zsv_ext1 = zsb + 1
				Else
					dx_ext1 = dx0 - 1 - SQUISH_CONSTANT_3D
					dy_ext1 = dy0 - 1 - SQUISH_CONSTANT_3D
					dz_ext1 = dz0 + 1 - SQUISH_CONSTANT_3D
					xsv_ext1 = xsb + 1
					ysv_ext1 = ysb + 1
					zsv_ext1 = zsb - 1
				EndIf
			EndIf
		Else ''One point on (0,0,0) side, one point on (1,1,1) side
			Dim As Byte c1, c2
			if (aIsFurtherSide) Then
				c1 = aPoint
				c2 = bPoint
			Else
				c1 = bPoint
				c2 = aPoint
			endif

			''One contribution is a permutation of (1,1,-1)
			if ((c1 and &h01) = 0) Then
				dx_ext0 = dx0 + 1 - SQUISH_CONSTANT_3D
				dy_ext0 = dy0 - 1 - SQUISH_CONSTANT_3D
				dz_ext0 = dz0 - 1 - SQUISH_CONSTANT_3D
				xsv_ext0 = xsb - 1
				ysv_ext0 = ysb + 1
				zsv_ext0 = zsb + 1
			elseif ((c1 and &h02) = 0) Then
				dx_ext0 = dx0 - 1 - SQUISH_CONSTANT_3D
				dy_ext0 = dy0 + 1 - SQUISH_CONSTANT_3D
				dz_ext0 = dz0 - 1 - SQUISH_CONSTANT_3D
				xsv_ext0 = xsb + 1
				ysv_ext0 = ysb - 1
				zsv_ext0 = zsb + 1
			Else
				dx_ext0 = dx0 - 1 - SQUISH_CONSTANT_3D
				dy_ext0 = dy0 - 1 - SQUISH_CONSTANT_3D
				dz_ext0 = dz0 + 1 - SQUISH_CONSTANT_3D
				xsv_ext0 = xsb + 1
				ysv_ext0 = ysb + 1
				zsv_ext0 = zsb - 1
			endif

			''One contribution is a permutation of (0,0,2)
			dx_ext1 = dx0 - 2 * SQUISH_CONSTANT_3D
			dy_ext1 = dy0 - 2 * SQUISH_CONSTANT_3D
			dz_ext1 = dz0 - 2 * SQUISH_CONSTANT_3D
			xsv_ext1 = xsb
			ysv_ext1 = ysb
			zsv_ext1 = zsb
			if ((c2 and &h01) <> 0) Then
				dx_ext1 -= 2
				xsv_ext1 += 2
			elseif ((c2 and &h02) <> 0) Then
				dy_ext1 -= 2
				ysv_ext1 += 2
			Else
				dz_ext1 -= 2
				zsv_ext1 += 2
			EndIf
		endif

		''Contribution (1,0,0)
		Dim As Double dx1 = dx0 - 1 - SQUISH_CONSTANT_3D
		Dim As Double dy1 = dy0 - 0 - SQUISH_CONSTANT_3D
		Dim As Double dz1 = dz0 - 0 - SQUISH_CONSTANT_3D
		Dim As Double attn1 = 2 - dx1 * dx1 - dy1 * dy1 - dz1 * dz1
		if (attn1 > 0) Then
			attn1 *= attn1
			value += attn1 * attn1 * extrapolate3(xsb + 1, ysb + 0, zsb + 0, dx1, dy1, dz1)
		EndIf

		''Contribution (0,1,0)
		Dim As Double dx2 = dx0 - 0 - SQUISH_CONSTANT_3D
		Dim As Double dy2 = dy0 - 1 - SQUISH_CONSTANT_3D
		Dim As Double dz2 = dz1
		Dim As Double attn2 = 2 - dx2 * dx2 - dy2 * dy2 - dz2 * dz2
		if (attn2 > 0) Then
			attn2 *= attn2
			value += attn2 * attn2 * extrapolate3(xsb + 0, ysb + 1, zsb + 0, dx2, dy2, dz2)
		EndIf

		''Contribution (0,0,1)
		Dim As Double dx3 = dx2
		Dim As Double dy3 = dy1
		Dim As Double dz3 = dz0 - 1 - SQUISH_CONSTANT_3D
		Dim As Double attn3 = 2 - dx3 * dx3 - dy3 * dy3 - dz3 * dz3
		if (attn3 > 0) Then
			attn3 *= attn3
			value += attn3 * attn3 * extrapolate3(xsb + 0, ysb + 0, zsb + 1, dx3, dy3, dz3)
		EndIf

		''Contribution (1,1,0)
		Dim As Double dx4 = dx0 - 1 - 2 * SQUISH_CONSTANT_3D
		Dim As Double dy4 = dy0 - 1 - 2 * SQUISH_CONSTANT_3D
		Dim As Double dz4 = dz0 - 0 - 2 * SQUISH_CONSTANT_3D
		Dim As Double attn4 = 2 - dx4 * dx4 - dy4 * dy4 - dz4 * dz4
		if (attn4 > 0) Then
			attn4 *= attn4
			value += attn4 * attn4 * extrapolate3(xsb + 1, ysb + 1, zsb + 0, dx4, dy4, dz4)
		EndIf

		''Contribution (1,0,1)
		Dim As Double dx5 = dx4
		Dim As Double dy5 = dy0 - 0 - 2 * SQUISH_CONSTANT_3D
		Dim As Double dz5 = dz0 - 1 - 2 * SQUISH_CONSTANT_3D
		Dim As Double attn5 = 2 - dx5 * dx5 - dy5 * dy5 - dz5 * dz5
		if (attn5 > 0) Then
			attn5 *= attn5
			value += attn5 * attn5 * extrapolate3(xsb + 1, ysb + 0, zsb + 1, dx5, dy5, dz5)
		EndIf

		''Contribution (0,1,1)
		Dim As Double dx6 = dx0 - 0 - 2 * SQUISH_CONSTANT_3D
		Dim As Double dy6 = dy4
		Dim As Double dz6 = dz5
		Dim As Double attn6 = 2 - dx6 * dx6 - dy6 * dy6 - dz6 * dz6
		if (attn6 > 0) Then
			attn6 *= attn6
			value += attn6 * attn6 * extrapolate3(xsb + 0, ysb + 1, zsb + 1, dx6, dy6, dz6)
		endif
	EndIf

	''First extra vertex
	Dim As Double attn_ext0 = 2 - dx_ext0 * dx_ext0 - dy_ext0 * dy_ext0 - dz_ext0 * dz_ext0
	if (attn_ext0 > 0) then
		attn_ext0 *= attn_ext0
		value += attn_ext0 * attn_ext0 * extrapolate3(xsv_ext0, ysv_ext0, zsv_ext0, dx_ext0, dy_ext0, dz_ext0)
	EndIf

	''Second extra vertex
	Dim As Double attn_ext1 = 2 - dx_ext1 * dx_ext1 - dy_ext1 * dy_ext1 - dz_ext1 * dz_ext1
	if (attn_ext1 > 0) Then
		attn_ext1 *= attn_ext1
		value += attn_ext1 * attn_ext1 * extrapolate3(xsv_ext1, ysv_ext1, zsv_ext1, dx_ext1, dy_ext1, dz_ext1)
	EndIf
	value/=NORM_CONSTANT_3D
	return value 
End Function

''4D OpenSimplex Noise.
Function OpenSimplexNoise.noise(x As Double ,y As double, z As Double, w As Double) As Double

	''Place input coordinates on simplectic honeycomb.
	Dim As Double stretchOffset = (x + y + z + w) * STRETCH_CONSTANT_4D
	Dim As Double xs = x + stretchOffset
	Dim As Double ys = y + stretchOffset
	Dim As Double zs = z + stretchOffset
	Dim As Double ws = w + stretchOffset

	''Floor to get simplectic honeycomb coordinates of rhombo-hypercube super-cell origin.
	Dim As Integer xsb = fastFloor(xs)
	Dim As Integer ysb = fastFloor(ys)
	Dim As Integer zsb = fastFloor(zs)
	Dim As Integer wsb = fastFloor(ws)

	''Skew out to get actual coordinates of stretched rhombo-hypercube origin. We'll need these later.
	Dim As Double squishOffset = (xsb + ysb + zsb + wsb) * SQUISH_CONSTANT_4D
	Dim As Double xb = xsb + squishOffset
	Dim As Double yb = ysb + squishOffset
	Dim As Double zb = zsb + squishOffset
	Dim As Double wb = wsb + squishOffset

	''Compute simplectic honeycomb coordinates relative to rhombo-hypercube origin.
	Dim As Double xins = xs - xsb
	Dim As Double yins = ys - ysb
	Dim As Double zins = zs - zsb
	Dim As Double wins = ws - wsb

	''Sum those together to get a value that determines which region we're in.
	Dim As Double inSum = xins + yins + zins + wins

	''Positions relative to origin point.
	Dim As Double dx0 = x - xb
	Dim As Double dy0 = y - yb
	Dim As Double dz0 = z - zb
	Dim As Double dw0 = w - wb

	''We'll be defining these inside the next block and using them afterwards.
	Dim As Double dx_ext0, dy_ext0, dz_ext0, dw_ext0
	Dim As Double dx_ext1, dy_ext1, dz_ext1, dw_ext1
	Dim As Double dx_ext2, dy_ext2, dz_ext2, dw_ext2
	Dim As Integer xsv_ext0, ysv_ext0, zsv_ext0, wsv_ext0
	Dim As Integer xsv_ext1, ysv_ext1, zsv_ext1, wsv_ext1
	Dim As Integer xsv_ext2, ysv_ext2, zsv_ext2, wsv_ext2

	Dim As Double value = 0
	if (inSum <= 1) Then ''We're inside the pentachoron (4-Simplex) at (0,0,0,0)

		''Determine which two of (0,0,0,1), (0,0,1,0), (0,1,0,0), (1,0,0,0) are closest.
		Dim As Byte aPoint = &h01
		Dim As Double aScore = xins
		Dim As Byte bPoint = &h02
		Dim As Double bScore = yins
		if (aScore >= bScore And zins > bScore) Then
			bScore = zins
			bPoint = &h04
		elseif (aScore < bScore And zins > aScore) Then
			aScore = zins
			aPoint = &h04
		EndIf
		if (aScore >= bScore And wins > bScore) Then
			bScore = wins
			bPoint = &h08
		ElseIf (aScore < bScore And wins > aScore) then
			aScore = wins
			aPoint = &h08
		endif

		''Now we determine the three lattice points not part of the pentachoron that may contribute.
		''This depends on the closest two pentachoron vertices, including (0,0,0,0)
		Dim As Double uins = 1 - inSum
		if (uins > aScore Or uins > bScore) Then ''(0,0,0,0) is one of the closest two pentachoron vertices.
			Dim As Byte c = iif(bScore > aScore , bPoint , aPoint) ''Our other closest vertex is the closest out of a and b.
			if ((c and &h01) = 0) Then
				xsv_ext0 = xsb - 1
				xsv_ext1 = xsv_ext2 = xsb
				dx_ext0 = dx0 + 1
				dx_ext1 = dx_ext2 = dx0
			Else
				xsv_ext0 = xsv_ext1 = xsv_ext2 = xsb + 1
				dx_ext0 = dx_ext1 = dx_ext2 = dx0 - 1
			EndIf

			if ((c and &h02) = 0) Then
				ysv_ext0 = ysv_ext1 = ysv_ext2 = ysb
				dy_ext0 = dy_ext1 = dy_ext2 = dy0
				if ((c and &h01) = &h01) Then
					ysv_ext0 -= 1
					dy_ext0 += 1
				Else
					ysv_ext1 -= 1
					dy_ext1 += 1
				EndIf
			Else
				ysv_ext0 = ysv_ext1 = ysv_ext2 = ysb + 1
				dy_ext0 = dy_ext1 = dy_ext2 = dy0 - 1
			EndIf

			if ((c and &h04) = 0) Then
				zsv_ext0 = zsv_ext1 = zsv_ext2 = zsb
				dz_ext0 = dz_ext1 = dz_ext2 = dz0
				if ((c and &h03) <> 0) Then
					if ((c and &h03) = &h03) Then
						zsv_ext0 -= 1
						dz_ext0 += 1
					Else
						zsv_ext1 -= 1
						dz_ext1 += 1
					EndIf
				Else
					zsv_ext2 -= 1
					dz_ext2 += 1
				EndIf
			Else
				zsv_ext0 = zsv_ext1 = zsv_ext2 = zsb + 1
				dz_ext0 = dz_ext1 = dz_ext2 = dz0 - 1
			EndIf

			if ((c and &h08) = 0) Then
				wsv_ext0 = wsv_ext1 = wsb
				wsv_ext2 = wsb - 1
				dw_ext0 = dw_ext1 = dw0
				dw_ext2 = dw0 + 1
			Else
				wsv_ext0 = wsv_ext1 = wsv_ext2 = wsb + 1
				dw_ext0 = dw_ext1 = dw_ext2 = dw0 - 1
			EndIf
		Else ''(0,0,0,0) is not one of the closest two pentachoron vertices.
			Dim As Byte c = cast(Byte,(aPoint or bPoint)) ''Our three extra vertices are determined by the closest two.

			if ((c and &h01) = 0) Then
				xsv_ext0 = xsv_ext2 = xsb
				xsv_ext1 = xsb - 1
				dx_ext0 = dx0 - 2 * SQUISH_CONSTANT_4D
				dx_ext1 = dx0 + 1 - SQUISH_CONSTANT_4D
				dx_ext2 = dx0 - SQUISH_CONSTANT_4D
			Else
				xsv_ext0 = xsv_ext1 = xsv_ext2 = xsb + 1
				dx_ext0 = dx0 - 1 - 2 * SQUISH_CONSTANT_4D
				dx_ext1 = dx_ext2 = dx0 - 1 - SQUISH_CONSTANT_4D
			EndIf

			if ((c and &h02) = 0) Then
				ysv_ext0 = ysv_ext1 = ysv_ext2 = ysb
				dy_ext0 = dy0 - 2 * SQUISH_CONSTANT_4D
				dy_ext1 = dy_ext2 = dy0 - SQUISH_CONSTANT_4D
				if ((c and &h01) = &h01) Then
					ysv_ext1 -= 1
					dy_ext1 += 1
				Else
					ysv_ext2 -= 1
					dy_ext2 += 1
				EndIf
			Else
				ysv_ext0 = ysv_ext1 = ysv_ext2 = ysb + 1
				dy_ext0 = dy0 - 1 - 2 * SQUISH_CONSTANT_4D
				dy_ext1 = dy_ext2 = dy0 - 1 - SQUISH_CONSTANT_4D
			EndIf

			if ((c and &h04) = 0) Then
				zsv_ext0 = zsv_ext1 = zsv_ext2 = zsb
				dz_ext0 = dz0 - 2 * SQUISH_CONSTANT_4D
				dz_ext1 = dz_ext2 = dz0 - SQUISH_CONSTANT_4D
				if ((c and &h03) = &h03) Then
					zsv_ext1 -= 1
					dz_ext1 += 1
				Else
					zsv_ext2 -= 1
					dz_ext2 += 1
				EndIf
			Else
				zsv_ext0 = zsv_ext1 = zsv_ext2 = zsb + 1
				dz_ext0 = dz0 - 1 - 2 * SQUISH_CONSTANT_4D
				dz_ext1 = dz_ext2 = dz0 - 1 - SQUISH_CONSTANT_4D
			EndIf

			if ((c and &h08) = 0) Then
				wsv_ext0 = wsv_ext1 = wsb
				wsv_ext2 = wsb - 1
				dw_ext0 = dw0 - 2 * SQUISH_CONSTANT_4D
				dw_ext1 = dw0 - SQUISH_CONSTANT_4D
				dw_ext2 = dw0 + 1 - SQUISH_CONSTANT_4D
			Else
				wsv_ext0 = wsv_ext1 = wsv_ext2 = wsb + 1
				dw_ext0 = dw0 - 1 - 2 * SQUISH_CONSTANT_4D
				dw_ext1 = dw_ext2 = dw0 - 1 - SQUISH_CONSTANT_4D
			EndIf
		endif

		''Contribution (0,0,0,0)
		Dim As Double attn0 = 2 - dx0 * dx0 - dy0 * dy0 - dz0 * dz0 - dw0 * dw0
		if (attn0 > 0) Then
			attn0 *= attn0
			value += attn0 * attn0 * extrapolate4(xsb + 0, ysb + 0, zsb + 0, wsb + 0, dx0, dy0, dz0, dw0)
		EndIf

		''Contribution (1,0,0,0)
		Dim As Double dx1 = dx0 - 1 - SQUISH_CONSTANT_4D
		Dim As Double dy1 = dy0 - 0 - SQUISH_CONSTANT_4D
		Dim As Double dz1 = dz0 - 0 - SQUISH_CONSTANT_4D
		Dim As Double dw1 = dw0 - 0 - SQUISH_CONSTANT_4D
		Dim As Double attn1 = 2 - dx1 * dx1 - dy1 * dy1 - dz1 * dz1 - dw1 * dw1
		if (attn1 > 0) Then
			attn1 *= attn1
			value += attn1 * attn1 * extrapolate4(xsb + 1, ysb + 0, zsb + 0, wsb + 0, dx1, dy1, dz1, dw1)
		EndIf

		''Contribution (0,1,0,0)
		Dim As Double dx2 = dx0 - 0 - SQUISH_CONSTANT_4D
		Dim As Double dy2 = dy0 - 1 - SQUISH_CONSTANT_4D
		Dim As Double dz2 = dz1
		Dim As Double dw2 = dw1
		Dim As Double attn2 = 2 - dx2 * dx2 - dy2 * dy2 - dz2 * dz2 - dw2 * dw2
		if (attn2 > 0) Then
			attn2 *= attn2
			value += attn2 * attn2 * extrapolate4(xsb + 0, ysb + 1, zsb + 0, wsb + 0, dx2, dy2, dz2, dw2)
		EndIf

		''Contribution (0,0,1,0)
		Dim As Double dx3 = dx2
		Dim As Double dy3 = dy1
		Dim As Double dz3 = dz0 - 1 - SQUISH_CONSTANT_4D
		Dim As Double dw3 = dw1
		Dim As Double attn3 = 2 - dx3 * dx3 - dy3 * dy3 - dz3 * dz3 - dw3 * dw3
		if (attn3 > 0) Then
			attn3 *= attn3
			value += attn3 * attn3 * extrapolate4(xsb + 0, ysb + 0, zsb + 1, wsb + 0, dx3, dy3, dz3, dw3)
		EndIf

		''Contribution (0,0,0,1)
		Dim As Double dx4 = dx2
		Dim As Double dy4 = dy1
		Dim As Double dz4 = dz1
		Dim As Double dw4 = dw0 - 1 - SQUISH_CONSTANT_4D
		Dim As Double attn4 = 2 - dx4 * dx4 - dy4 * dy4 - dz4 * dz4 - dw4 * dw4
		if (attn4 > 0) Then
			attn4 *= attn4
			value += attn4 * attn4 * extrapolate4(xsb + 0, ysb + 0, zsb + 0, wsb + 1, dx4, dy4, dz4, dw4)
		EndIf
	Elseif (inSum >= 3) Then ''We're inside the pentachoron (4-Simplex) at (1,1,1,1)
		''Determine which two of (1,1,1,0), (1,1,0,1), (1,0,1,1), (0,1,1,1) are closest.
		Dim As Byte aPoint = &h0E
		Dim As Double aScore = xins
		Dim As Byte bPoint = &h0D
		Dim As Double bScore = yins
		if (aScore <= bScore And zins < bScore) Then
			bScore = zins
			bPoint = &h0B
		ElseIf (aScore > bScore And zins < aScore) then
			aScore = zins
			aPoint = &h0B
		EndIf
		if (aScore <= bScore And wins < bScore) Then
			bScore = wins
			bPoint = &h07
		ElseIf (aScore > bScore And wins < aScore) then
			aScore = wins
			aPoint = &h07
		EndIf

		''Now we determine the three lattice points not part of the pentachoron that may contribute.
		''This depends on the closest two pentachoron vertices, including (0,0,0,0)
		Dim As Double uins = 4 - inSum
		if (uins < aScore Or uins < bScore) Then ''(1,1,1,1) is one of the closest two pentachoron vertices.
			Dim As Byte c = iif(bScore < aScore , bPoint , aPoint) ''Our other closest vertex is the closest out of a and b.

			if ((c and &h01) <> 0) Then
				xsv_ext0 = xsb + 2
				xsv_ext1 = xsv_ext2 = xsb + 1
				dx_ext0 = dx0 - 2 - 4 * SQUISH_CONSTANT_4D
				dx_ext1 = dx_ext2 = dx0 - 1 - 4 * SQUISH_CONSTANT_4D
			Else
				xsv_ext0 = xsv_ext1 = xsv_ext2 = xsb
				dx_ext0 = dx_ext1 = dx_ext2 = dx0 - 4 * SQUISH_CONSTANT_4D
			EndIf

			if ((c and &h02) <> 0) Then
				ysv_ext0 = ysv_ext1 = ysv_ext2 = ysb + 1
				dy_ext0 = dy_ext1 = dy_ext2 = dy0 - 1 - 4 * SQUISH_CONSTANT_4D
				if ((c and &h01) <> 0) Then
					ysv_ext1 += 1
					dy_ext1 -= 1
				Else
					ysv_ext0 += 1
					dy_ext0 -= 1
				EndIf
			Else
				ysv_ext0 = ysv_ext1 = ysv_ext2 = ysb
				dy_ext0 = dy_ext1 = dy_ext2 = dy0 - 4 * SQUISH_CONSTANT_4D
			EndIf

			if ((c and &h04) <> 0) Then
				zsv_ext0 = zsv_ext1 = zsv_ext2 = zsb + 1
				dz_ext0 = dz_ext1 = dz_ext2 = dz0 - 1 - 4 * SQUISH_CONSTANT_4D
				if ((c and &h03) <> &h03) Then
					if ((c and &h03) = 0) Then
						zsv_ext0 += 1
						dz_ext0 -= 1
					Else
						zsv_ext1 += 1
						dz_ext1 -= 1
					EndIf
				Else
					zsv_ext2 += 1
					dz_ext2 -= 1
				EndIf
			Else
				zsv_ext0 = zsv_ext1 = zsv_ext2 = zsb
				dz_ext0 = dz_ext1 = dz_ext2 = dz0 - 4 * SQUISH_CONSTANT_4D
			EndIf

			if ((c and &h08) <> 0) Then
				wsv_ext0 = wsv_ext1 = wsb + 1
				wsv_ext2 = wsb + 2
				dw_ext0 = dw_ext1 = dw0 - 1 - 4 * SQUISH_CONSTANT_4D
				dw_ext2 = dw0 - 2 - 4 * SQUISH_CONSTANT_4D
			Else
				wsv_ext0 = wsv_ext1 = wsv_ext2 = wsb
				dw_ext0 = dw_ext1 = dw_ext2 = dw0 - 4 * SQUISH_CONSTANT_4D
			EndIf
		Else ''(1,1,1,1) is not one of the closest two pentachoron vertices.
			Dim As Byte c = cast(Byte,(aPoint And bPoint)) ''Our three extra vertices are determined by the closest two.

			if ((c and &h01) <> 0) Then
				xsv_ext0 = xsv_ext2 = xsb + 1
				xsv_ext1 = xsb + 2
				dx_ext0 = dx0 - 1 - 2 * SQUISH_CONSTANT_4D
				dx_ext1 = dx0 - 2 - 3 * SQUISH_CONSTANT_4D
				dx_ext2 = dx0 - 1 - 3 * SQUISH_CONSTANT_4D
			Else
				xsv_ext0 = xsv_ext1 = xsv_ext2 = xsb
				dx_ext0 = dx0 - 2 * SQUISH_CONSTANT_4D
				dx_ext1 = dx_ext2 = dx0 - 3 * SQUISH_CONSTANT_4D
			EndIf

			if ((c and &h02) <> 0) Then
				ysv_ext0 = ysv_ext1 = ysv_ext2 = ysb + 1
				dy_ext0 = dy0 - 1 - 2 * SQUISH_CONSTANT_4D
				dy_ext1 = dy_ext2 = dy0 - 1 - 3 * SQUISH_CONSTANT_4D
				if ((c and &h01) <> 0) Then
					ysv_ext2 += 1
					dy_ext2 -= 1
				Else
					ysv_ext1 += 1
					dy_ext1 -= 1
				EndIf
			Else
				ysv_ext0 = ysv_ext1 = ysv_ext2 = ysb
				dy_ext0 = dy0 - 2 * SQUISH_CONSTANT_4D
				dy_ext1 = dy_ext2 = dy0 - 3 * SQUISH_CONSTANT_4D
			EndIf

			if ((c and &h04) <> 0) Then
				zsv_ext0 = zsv_ext1 = zsv_ext2 = zsb + 1
				dz_ext0 = dz0 - 1 - 2 * SQUISH_CONSTANT_4D
				dz_ext1 = dz_ext2 = dz0 - 1 - 3 * SQUISH_CONSTANT_4D
				if ((c and &h03) <> 0) Then
					zsv_ext2 += 1
					dz_ext2 -= 1
				Else
					zsv_ext1 += 1
					dz_ext1 -= 1
				EndIf
			Else
				zsv_ext0 = zsv_ext1 = zsv_ext2 = zsb
				dz_ext0 = dz0 - 2 * SQUISH_CONSTANT_4D
				dz_ext1 = dz_ext2 = dz0 - 3 * SQUISH_CONSTANT_4D
			EndIf

			if ((c and &h08) <> 0) Then
				wsv_ext0 = wsv_ext1 = wsb + 1
				wsv_ext2 = wsb + 2
				dw_ext0 = dw0 - 1 - 2 * SQUISH_CONSTANT_4D
				dw_ext1 = dw0 - 1 - 3 * SQUISH_CONSTANT_4D
				dw_ext2 = dw0 - 2 - 3 * SQUISH_CONSTANT_4D
			Else
				wsv_ext0 = wsv_ext1 = wsv_ext2 = wsb
				dw_ext0 = dw0 - 2 * SQUISH_CONSTANT_4D
				dw_ext1 = dw_ext2 = dw0 - 3 * SQUISH_CONSTANT_4D
			EndIf
		endif

		''Contribution (1,1,1,0)
		Dim As Double dx4 = dx0 - 1 - 3 * SQUISH_CONSTANT_4D
		Dim As Double dy4 = dy0 - 1 - 3 * SQUISH_CONSTANT_4D
		Dim As Double dz4 = dz0 - 1 - 3 * SQUISH_CONSTANT_4D
		Dim As Double dw4 = dw0 - 3 * SQUISH_CONSTANT_4D
		Dim As Double attn4 = 2 - dx4 * dx4 - dy4 * dy4 - dz4 * dz4 - dw4 * dw4
		if (attn4 > 0) Then
			attn4 *= attn4
			value += attn4 * attn4 * extrapolate4(xsb + 1, ysb + 1, zsb + 1, wsb + 0, dx4, dy4, dz4, dw4)
		EndIf

		''Contribution (1,1,0,1)
		Dim As Double dx3 = dx4
		Dim As Double dy3 = dy4
		Dim As Double dz3 = dz0 - 3 * SQUISH_CONSTANT_4D
		Dim As Double dw3 = dw0 - 1 - 3 * SQUISH_CONSTANT_4D
		Dim As Double attn3 = 2 - dx3 * dx3 - dy3 * dy3 - dz3 * dz3 - dw3 * dw3
		if (attn3 > 0) Then
			attn3 *= attn3
			value += attn3 * attn3 * extrapolate4(xsb + 1, ysb + 1, zsb + 0, wsb + 1, dx3, dy3, dz3, dw3)
		EndIf

		''Contribution (1,0,1,1)
		Dim As Double dx2 = dx4
		Dim As Double dy2 = dy0 - 3 * SQUISH_CONSTANT_4D
		Dim As Double dz2 = dz4
		Dim As Double dw2 = dw3
		Dim As Double attn2 = 2 - dx2 * dx2 - dy2 * dy2 - dz2 * dz2 - dw2 * dw2
		if (attn2 > 0) Then
			attn2 *= attn2
			value += attn2 * attn2 * extrapolate4(xsb + 1, ysb + 0, zsb + 1, wsb + 1, dx2, dy2, dz2, dw2)
		EndIf

		''Contribution (0,1,1,1)
		Dim As Double dx1 = dx0 - 3 * SQUISH_CONSTANT_4D
		Dim As Double dz1 = dz4
		Dim As Double dy1 = dy4
		Dim As Double dw1 = dw3
		Dim As Double attn1 = 2 - dx1 * dx1 - dy1 * dy1 - dz1 * dz1 - dw1 * dw1
		if (attn1 > 0) Then
			attn1 *= attn1
			value += attn1 * attn1 * extrapolate4(xsb + 0, ysb + 1, zsb + 1, wsb + 1, dx1, dy1, dz1, dw1)
		EndIf

		''Contribution (1,1,1,1)
		dx0 = dx0 - 1 - 4 * SQUISH_CONSTANT_4D
		dy0 = dy0 - 1 - 4 * SQUISH_CONSTANT_4D
		dz0 = dz0 - 1 - 4 * SQUISH_CONSTANT_4D
		dw0 = dw0 - 1 - 4 * SQUISH_CONSTANT_4D
		Dim As Double attn0 = 2 - dx0 * dx0 - dy0 * dy0 - dz0 * dz0 - dw0 * dw0
		if (attn0 > 0) Then
			attn0 *= attn0
			value += attn0 * attn0 * extrapolate4(xsb + 1, ysb + 1, zsb + 1, wsb + 1, dx0, dy0, dz0, dw0)
		EndIf
	ElseIf (inSum <= 2) then ''We're inside the first dispentachoron (Rectified 4-Simplex)
		Dim As Double aScore
		Dim As Byte aPoint
		Dim As boolean aIsBiggerSide = TRUE
		Dim As Double bScore
		Dim As Byte bPoint
		Dim As boolean bIsBiggerSide = TRUE

		''Decide between (1,1,0,0) and (0,0,1,1)
		if (xins + yins > zins + wins) Then
			aScore = xins + yins
			aPoint = &h03
		Else
			aScore = zins + wins
			aPoint = &h0C
		EndIf

		''Decide between (1,0,1,0) and (0,1,0,1)
		if (xins + zins > yins + wins) Then
			bScore = xins + zins
			bPoint = &h05
		Else
			bScore = yins + wins
			bPoint = &h0A
		EndIf

		''Closer between (1,0,0,1) and (0,1,1,0) will replace the further of a and b, if closer.
		if (xins + wins > yins + zins) Then
			Dim As Double score = xins + wins
			if (aScore >= bScore And score > bScore) Then
				bScore = score
				bPoint = &h09
			ElseIf (aScore < bScore And score > aScore) then
				aScore = score
				aPoint = &h09
			EndIf
		Else
			Dim As Double score = yins + zins
			if (aScore >= bScore And score > bScore) Then
				bScore = score
				bPoint = &h06
			ElseIf (aScore < bScore And score > aScore) then
				aScore = score
				aPoint = &h06
			EndIf
		EndIf

		''Decide if (1,0,0,0) is closer.
		Dim As Double p1 = 2 - inSum + xins
		if (aScore >= bScore And p1 > bScore) Then
			bScore = p1
			bPoint = &h01
			bIsBiggerSide = FALSE
		ElseIf (aScore < bScore And p1 > aScore) then
			aScore = p1
			aPoint = &h01
			aIsBiggerSide = FALSE
		EndIf

		''Decide if (0,1,0,0) is closer.
		Dim As Double p2 = 2 - inSum + yins
		if (aScore >= bScore And p2 > bScore) Then
			bScore = p2
			bPoint = &h02
			bIsBiggerSide = FALSE
		ElseIf (aScore < bScore And p2 > aScore) Then
			aScore = p2
			aPoint = &h02
			aIsBiggerSide = FALSE
		EndIf

		''Decide if (0,0,1,0) is closer.
		Dim As Double p3 = 2 - inSum + zins
		if (aScore >= bScore And p3 > bScore) Then
			bScore = p3
			bPoint = &h04
			bIsBiggerSide = FALSE
		elseif (aScore < bScore And p3 > aScore) Then
			aScore = p3
			aPoint = &h04
			aIsBiggerSide = FALSE
		EndIf

		''Decide if (0,0,0,1) is closer.
		Dim As Double p4 = 2 - inSum + wins
		if (aScore >= bScore And p4 > bScore) Then
			bScore = p4
			bPoint = &h08
			bIsBiggerSide = FALSE
		ElseIf (aScore < bScore And p4 > aScore) then
			aScore = p4
			aPoint = &h08
			aIsBiggerSide = FALSE
		endif

		''Where each of the two closest points are determines how the extra three vertices are calculated.
		if (aIsBiggerSide = bIsBiggerSide) Then
			if (aIsBiggerSide) then ''Both closest points on the bigger side
				Dim As Byte c1 = Cast(Byte,(aPoint or bPoint))
				Dim As Byte c2 = cast(Byte,(aPoint And bPoint))
				if ((c1 and &h01) = 0) Then
					xsv_ext0 = xsb
					xsv_ext1 = xsb - 1
					dx_ext0 = dx0 - 3 * SQUISH_CONSTANT_4D
					dx_ext1 = dx0 + 1 - 2 * SQUISH_CONSTANT_4D
				Else
					xsv_ext0 = xsv_ext1 = xsb + 1
					dx_ext0 = dx0 - 1 - 3 * SQUISH_CONSTANT_4D
					dx_ext1 = dx0 - 1 - 2 * SQUISH_CONSTANT_4D
				EndIf

				if ((c1 and &h02) = 0) Then
					ysv_ext0 = ysb
					ysv_ext1 = ysb - 1
					dy_ext0 = dy0 - 3 * SQUISH_CONSTANT_4D
					dy_ext1 = dy0 + 1 - 2 * SQUISH_CONSTANT_4D
				Else
					ysv_ext0 = ysv_ext1 = ysb + 1
					dy_ext0 = dy0 - 1 - 3 * SQUISH_CONSTANT_4D
					dy_ext1 = dy0 - 1 - 2 * SQUISH_CONSTANT_4D
				endif

				if ((c1 and &h04) = 0) Then
					zsv_ext0 = zsb
					zsv_ext1 = zsb - 1
					dz_ext0 = dz0 - 3 * SQUISH_CONSTANT_4D
					dz_ext1 = dz0 + 1 - 2 * SQUISH_CONSTANT_4D
				Else
					zsv_ext0 = zsv_ext1 = zsb + 1
					dz_ext0 = dz0 - 1 - 3 * SQUISH_CONSTANT_4D
					dz_ext1 = dz0 - 1 - 2 * SQUISH_CONSTANT_4D
				EndIf

				if ((c1 and &h08) = 0) Then
					wsv_ext0 = wsb
					wsv_ext1 = wsb - 1
					dw_ext0 = dw0 - 3 * SQUISH_CONSTANT_4D
					dw_ext1 = dw0 + 1 - 2 * SQUISH_CONSTANT_4D
				Else
					wsv_ext0 = wsv_ext1 = wsb + 1
					dw_ext0 = dw0 - 1 - 3 * SQUISH_CONSTANT_4D
					dw_ext1 = dw0 - 1 - 2 * SQUISH_CONSTANT_4D
				endif

				''One combination is a permutation of (0,0,0,2) based on c2
				xsv_ext2 = xsb
				ysv_ext2 = ysb
				zsv_ext2 = zsb
				wsv_ext2 = wsb
				dx_ext2 = dx0 - 2 * SQUISH_CONSTANT_4D
				dy_ext2 = dy0 - 2 * SQUISH_CONSTANT_4D
				dz_ext2 = dz0 - 2 * SQUISH_CONSTANT_4D
				dw_ext2 = dw0 - 2 * SQUISH_CONSTANT_4D
				if ((c2 and &h01) <> 0) Then
					xsv_ext2 += 2
					dx_ext2 -= 2
				ElseIf ((c2 and &h02) <> 0) Then
					ysv_ext2 += 2
					dy_ext2 -= 2
				ElseIf ((c2 and &h04) <> 0) then
					zsv_ext2 += 2
					dz_ext2 -= 2
				Else
					wsv_ext2 += 2
					dw_ext2 -= 2
				endif

			Else ''Both closest points on the smaller side
				''One of the two extra points is (0,0,0,0)
				xsv_ext2 = xsb
				ysv_ext2 = ysb
				zsv_ext2 = zsb
				wsv_ext2 = wsb
				dx_ext2 = dx0
				dy_ext2 = dy0
				dz_ext2 = dz0
				dw_ext2 = dw0

				''Other two points are based on the omitted axes.
				Dim As Byte c = cast(Byte,(aPoint or bPoint))

				if ((c and &h01) = 0) Then
					xsv_ext0 = xsb - 1
					xsv_ext1 = xsb
					dx_ext0 = dx0 + 1 - SQUISH_CONSTANT_4D
					dx_ext1 = dx0 - SQUISH_CONSTANT_4D
				Else
					xsv_ext0 = xsv_ext1 = xsb + 1
					dx_ext0 = dx_ext1 = dx0 - 1 - SQUISH_CONSTANT_4D
				EndIf

				if ((c and &h02) = 0) Then
					ysv_ext0 = ysv_ext1 = ysb
					dy_ext0 = dy_ext1 = dy0 - SQUISH_CONSTANT_4D
					if ((c and &h01) = &h01) Then
						ysv_ext0 -= 1
						dy_ext0 += 1
					Else
						ysv_ext1 -= 1
						dy_ext1 += 1
					EndIf
				Else
					ysv_ext0 = ysv_ext1 = ysb + 1
					dy_ext0 = dy_ext1 = dy0 - 1 - SQUISH_CONSTANT_4D
				EndIf

				if ((c and &h04) = 0) Then
					zsv_ext0 = zsv_ext1 = zsb
					dz_ext0 = dz_ext1 = dz0 - SQUISH_CONSTANT_4D
					if ((c and &h03) = &h03) Then
						zsv_ext0 -= 1
						dz_ext0 += 1
					Else
						zsv_ext1 -= 1
						dz_ext1 += 1
					EndIf
				Else
					zsv_ext0 = zsv_ext1 = zsb + 1
					dz_ext0 = dz_ext1 = dz0 - 1 - SQUISH_CONSTANT_4D
				EndIf

				if ((c and &h08) = 0) Then
					wsv_ext0 = wsb
					wsv_ext1 = wsb - 1
					dw_ext0 = dw0 - SQUISH_CONSTANT_4D
					dw_ext1 = dw0 + 1 - SQUISH_CONSTANT_4D
				Else
					wsv_ext0 = wsv_ext1 = wsb + 1
					dw_ext0 = dw_ext1 = dw0 - 1 - SQUISH_CONSTANT_4D
				EndIf

			EndIf
		Else ''One point on each "side"
			Dim As Byte c1, c2
			if (aIsBiggerSide) Then
				c1 = aPoint
				c2 = bPoint
			Else
				c1 = bPoint
				c2 = aPoint
			EndIf

			''Two contributions are the bigger-sided point with each 0 replaced with -1.
			if ((c1 and &h01) = 0) Then
				xsv_ext0 = xsb - 1
				xsv_ext1 = xsb
				dx_ext0 = dx0 + 1 - SQUISH_CONSTANT_4D
				dx_ext1 = dx0 - SQUISH_CONSTANT_4D
			Else
				xsv_ext0 = xsv_ext1 = xsb + 1
				dx_ext0 = dx_ext1 = dx0 - 1 - SQUISH_CONSTANT_4D
			EndIf

			if ((c1 and &h02) = 0) Then
				ysv_ext0 = ysv_ext1 = ysb
				dy_ext0 = dy_ext1 = dy0 - SQUISH_CONSTANT_4D
				if ((c1 and &h01) = &h01) Then
					ysv_ext0 -= 1
					dy_ext0 += 1
				Else
					ysv_ext1 -= 1
					dy_ext1 += 1
				EndIf
			Else
				ysv_ext0 = ysv_ext1 = ysb + 1
				dy_ext0 = dy_ext1 = dy0 - 1 - SQUISH_CONSTANT_4D
			EndIf

			if ((c1 and &h04) = 0) Then
				zsv_ext0 = zsv_ext1 = zsb
				dz_ext0 = dz_ext1 = dz0 - SQUISH_CONSTANT_4D
				if ((c1 and &h03) = &h03) Then
					zsv_ext0 -= 1
					dz_ext0 += 1
				Else
					zsv_ext1 -= 1
					dz_ext1 += 1
				EndIf
			Else
				zsv_ext0 = zsv_ext1 = zsb + 1
				dz_ext0 = dz_ext1 = dz0 - 1 - SQUISH_CONSTANT_4D
			EndIf

			if ((c1 and &h08) = 0) Then
				wsv_ext0 = wsb
				wsv_ext1 = wsb - 1
				dw_ext0 = dw0 - SQUISH_CONSTANT_4D
				dw_ext1 = dw0 + 1 - SQUISH_CONSTANT_4D
			Else
				wsv_ext0 = wsv_ext1 = wsb + 1
				dw_ext0 = dw_ext1 = dw0 - 1 - SQUISH_CONSTANT_4D
			EndIf

			''One contribution is a permutation of (0,0,0,2) based on the smaller-sided point
			xsv_ext2 = xsb
			ysv_ext2 = ysb
			zsv_ext2 = zsb
			wsv_ext2 = wsb
			dx_ext2 = dx0 - 2 * SQUISH_CONSTANT_4D
			dy_ext2 = dy0 - 2 * SQUISH_CONSTANT_4D
			dz_ext2 = dz0 - 2 * SQUISH_CONSTANT_4D
			dw_ext2 = dw0 - 2 * SQUISH_CONSTANT_4D
			if ((c2 and &h01) <> 0) Then
				xsv_ext2 += 2
				dx_ext2 -= 2
			ElseIf ((c2 and &h02) <> 0) Then
				ysv_ext2 += 2
				dy_ext2 -= 2
			elseif ((c2 and &h04) <> 0) Then
				zsv_ext2 += 2
				dz_ext2 -= 2
			Else
				wsv_ext2 += 2
				dw_ext2 -= 2
			EndIf
		EndIf

		''Contribution (1,0,0,0)
		Dim As Double dx1 = dx0 - 1 - SQUISH_CONSTANT_4D
		Dim As Double dy1 = dy0 - 0 - SQUISH_CONSTANT_4D
		Dim As Double dz1 = dz0 - 0 - SQUISH_CONSTANT_4D
		Dim As Double dw1 = dw0 - 0 - SQUISH_CONSTANT_4D
		Dim As Double attn1 = 2 - dx1 * dx1 - dy1 * dy1 - dz1 * dz1 - dw1 * dw1
		if (attn1 > 0) Then
			attn1 *= attn1
			value += attn1 * attn1 * extrapolate4(xsb + 1, ysb + 0, zsb + 0, wsb + 0, dx1, dy1, dz1, dw1)
		EndIf

		''Contribution (0,1,0,0)
		Dim As Double dx2 = dx0 - 0 - SQUISH_CONSTANT_4D
		Dim As Double dy2 = dy0 - 1 - SQUISH_CONSTANT_4D
		Dim As Double dz2 = dz1
		Dim As Double dw2 = dw1
		Dim As Double attn2 = 2 - dx2 * dx2 - dy2 * dy2 - dz2 * dz2 - dw2 * dw2
		if (attn2 > 0) Then
			attn2 *= attn2
			value += attn2 * attn2 * extrapolate4(xsb + 0, ysb + 1, zsb + 0, wsb + 0, dx2, dy2, dz2, dw2)
		EndIf

		''Contribution (0,0,1,0)
		Dim As Double dx3 = dx2
		Dim As Double dy3 = dy1
		Dim As Double dz3 = dz0 - 1 - SQUISH_CONSTANT_4D
		Dim As Double dw3 = dw1
		Dim As Double attn3 = 2 - dx3 * dx3 - dy3 * dy3 - dz3 * dz3 - dw3 * dw3
		if (attn3 > 0) Then
			attn3 *= attn3
			value += attn3 * attn3 * extrapolate4(xsb + 0, ysb + 0, zsb + 1, wsb + 0, dx3, dy3, dz3, dw3)
		EndIf

		''Contribution (0,0,0,1)
		Dim As Double dx4 = dx2
		Dim As Double dy4 = dy1
		Dim As Double dz4 = dz1
		Dim As Double dw4 = dw0 - 1 - SQUISH_CONSTANT_4D
		Dim As Double attn4 = 2 - dx4 * dx4 - dy4 * dy4 - dz4 * dz4 - dw4 * dw4
		if (attn4 > 0) Then
			attn4 *= attn4
			value += attn4 * attn4 * extrapolate4(xsb + 0, ysb + 0, zsb + 0, wsb + 1, dx4, dy4, dz4, dw4)
		EndIf

		''Contribution (1,1,0,0)
		Dim As Double dx5 = dx0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dy5 = dy0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dz5 = dz0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dw5 = dw0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double attn5 = 2 - dx5 * dx5 - dy5 * dy5 - dz5 * dz5 - dw5 * dw5
		if (attn5 > 0) Then
			attn5 *= attn5
			value += attn5 * attn5 * extrapolate4(xsb + 1, ysb + 1, zsb + 0, wsb + 0, dx5, dy5, dz5, dw5)
		EndIf

		''Contribution (1,0,1,0)
		Dim As Double dx6 = dx0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dy6 = dy0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dz6 = dz0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dw6 = dw0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double attn6 = 2 - dx6 * dx6 - dy6 * dy6 - dz6 * dz6 - dw6 * dw6
		if (attn6 > 0) Then
			attn6 *= attn6
			value += attn6 * attn6 * extrapolate4(xsb + 1, ysb + 0, zsb + 1, wsb + 0, dx6, dy6, dz6, dw6)
		EndIf

		''Contribution (1,0,0,1)
		Dim As Double dx7 = dx0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dy7 = dy0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dz7 = dz0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dw7 = dw0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double attn7 = 2 - dx7 * dx7 - dy7 * dy7 - dz7 * dz7 - dw7 * dw7
		if (attn7 > 0) Then
			attn7 *= attn7
			value += attn7 * attn7 * extrapolate4(xsb + 1, ysb + 0, zsb + 0, wsb + 1, dx7, dy7, dz7, dw7)
		EndIf

		''Contribution (0,1,1,0)
		Dim As Double dx8 = dx0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dy8 = dy0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dz8 = dz0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dw8 = dw0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double attn8 = 2 - dx8 * dx8 - dy8 * dy8 - dz8 * dz8 - dw8 * dw8
		if (attn8 > 0) Then
			attn8 *= attn8
			value += attn8 * attn8 * extrapolate4(xsb + 0, ysb + 1, zsb + 1, wsb + 0, dx8, dy8, dz8, dw8)
		EndIf

		''Contribution (0,1,0,1)
		Dim As Double dx9 = dx0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dy9 = dy0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dz9 = dz0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dw9 = dw0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double attn9 = 2 - dx9 * dx9 - dy9 * dy9 - dz9 * dz9 - dw9 * dw9
		if (attn9 > 0) Then
			attn9 *= attn9
			value += attn9 * attn9 * extrapolate4(xsb + 0, ysb + 1, zsb + 0, wsb + 1, dx9, dy9, dz9, dw9)
		EndIf

		''Contribution (0,0,1,1)
		Dim As Double dx10 = dx0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dy10 = dy0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dz10 = dz0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dw10 = dw0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double attn10 = 2 - dx10 * dx10 - dy10 * dy10 - dz10 * dz10 - dw10 * dw10
		if (attn10 > 0) Then
			attn10 *= attn10
			value += attn10 * attn10 * extrapolate4(xsb + 0, ysb + 0, zsb + 1, wsb + 1, dx10, dy10, dz10, dw10)
		EndIf
	Else ''We're inside the second dispentachoron (Rectified 4-Simplex)
		Dim As Double aScore
		Dim As Byte aPoint
		Dim As boolean aIsBiggerSide = TRUE
		Dim As Double bScore
		Dim As Byte bPoint
		Dim As boolean bIsBiggerSide = TRUE

		''Decide between (0,0,1,1) and (1,1,0,0)
		if (xins + yins < zins + wins) Then
			aScore = xins + yins
			aPoint = &h0C
		Else
			aScore = zins + wins
			aPoint = &h03
		EndIf

		''Decide between (0,1,0,1) and (1,0,1,0)
		if (xins + zins < yins + wins) Then
			bScore = xins + zins
			bPoint = &h0A
		Else
			bScore = yins + wins
			bPoint = &h05
		EndIf

		''Closer between (0,1,1,0) and (1,0,0,1) will replace the further of a and b, if closer.
		if (xins + wins < yins + zins) Then
			Dim As Double score = xins + wins
			if (aScore <= bScore And score < bScore) Then
				bScore = score
				bPoint = &h06
			ElseIf (aScore > bScore And score < aScore) Then
				aScore = score
				aPoint = &h06
			EndIf
		Else
			Dim As Double score = yins + zins
			if (aScore <= bScore And score < bScore) Then
				bScore = score
				bPoint = &h09
			elseif (aScore > bScore And score < aScore) Then
				aScore = score
				aPoint = &h09
			EndIf
		EndIf

		''Decide if (0,1,1,1) is closer.
		Dim As Double p1 = 3 - inSum + xins
		if (aScore <= bScore And p1 < bScore) Then
			bScore = p1
			bPoint = &h0E
			bIsBiggerSide = FALSE
		ElseIf (aScore > bScore And p1 < aScore) Then
			aScore = p1
			aPoint = &h0E
			aIsBiggerSide = FALSE
		EndIf

		''Decide if (1,0,1,1) is closer.
		Dim As Double p2 = 3 - inSum + yins
		if (aScore <= bScore And p2 < bScore) Then
			bScore = p2
			bPoint = &h0D
			bIsBiggerSide = FALSE
		ElseIf (aScore > bScore And p2 < aScore) Then
			aScore = p2
			aPoint = &h0D
			aIsBiggerSide = FALSE
		EndIf

		''Decide if (1,1,0,1) is closer.
		Dim As Double p3 = 3 - inSum + zins
		if (aScore <= bScore And p3 < bScore) Then
			bScore = p3
			bPoint = &h0B
			bIsBiggerSide = FALSE
		ElseIf (aScore > bScore And p3 < aScore) Then
			aScore = p3
			aPoint = &h0B
			aIsBiggerSide = FALSE
		EndIf

		''Decide if (1,1,1,0) is closer.
		Dim As Double p4 = 3 - inSum + wins
		if (aScore <= bScore And p4 < bScore) Then
			bScore = p4
			bPoint = &h07
			bIsBiggerSide = FALSE
		elseif (aScore > bScore and p4 < aScore) Then
			aScore = p4
			aPoint = &h07
			aIsBiggerSide = FALSE
		EndIf

		''Where each of the two closest points are determines how the extra three vertices are calculated.
		if (aIsBiggerSide = bIsBiggerSide) Then
			if (aIsBiggerSide) Then ''Both closest points on the bigger side
				Dim As Byte c1 = Cast(Byte,(aPoint And bPoint))
				Dim As Byte c2 = cast(Byte,(aPoint or bPoint))

				''Two contributions are permutations of (0,0,0,1) and (0,0,0,2) based on c1
				xsv_ext0 = xsv_ext1 = xsb
				ysv_ext0 = ysv_ext1 = ysb
				zsv_ext0 = zsv_ext1 = zsb
				wsv_ext0 = wsv_ext1 = wsb
				dx_ext0 = dx0 - SQUISH_CONSTANT_4D
				dy_ext0 = dy0 - SQUISH_CONSTANT_4D
				dz_ext0 = dz0 - SQUISH_CONSTANT_4D
				dw_ext0 = dw0 - SQUISH_CONSTANT_4D
				dx_ext1 = dx0 - 2 * SQUISH_CONSTANT_4D
				dy_ext1 = dy0 - 2 * SQUISH_CONSTANT_4D
				dz_ext1 = dz0 - 2 * SQUISH_CONSTANT_4D
				dw_ext1 = dw0 - 2 * SQUISH_CONSTANT_4D
				if ((c1 and &h01) <> 0) Then
					xsv_ext0 += 1
					dx_ext0 -= 1
					xsv_ext1 += 2
					dx_ext1 -= 2
				ElseIf ((c1 and &h02) <> 0) Then
					ysv_ext0 += 1
					dy_ext0 -= 1
					ysv_ext1 += 2
					dy_ext1 -= 2
				ElseIf ((c1 and &h04) <> 0) Then
					zsv_ext0 += 1
					dz_ext0 -= 1
					zsv_ext1 += 2
					dz_ext1 -= 2
				Else
					wsv_ext0 += 1
					dw_ext0 -= 1
					wsv_ext1 += 2
					dw_ext1 -= 2
				EndIf

				''One contribution is a permutation of (1,1,1,-1) based on c2
				xsv_ext2 = xsb + 1
				ysv_ext2 = ysb + 1
				zsv_ext2 = zsb + 1
				wsv_ext2 = wsb + 1
				dx_ext2 = dx0 - 1 - 2 * SQUISH_CONSTANT_4D
				dy_ext2 = dy0 - 1 - 2 * SQUISH_CONSTANT_4D
				dz_ext2 = dz0 - 1 - 2 * SQUISH_CONSTANT_4D
				dw_ext2 = dw0 - 1 - 2 * SQUISH_CONSTANT_4D
				if ((c2 and &h01) = 0) Then
					xsv_ext2 -= 2
					dx_ext2 += 2
				ElseIf ((c2 and &h02) = 0) Then
					ysv_ext2 -= 2
					dy_ext2 += 2
				ElseIf ((c2 and &h04) = 0) Then
					zsv_ext2 -= 2
					dz_ext2 += 2
				Else
					wsv_ext2 -= 2
					dw_ext2 += 2
				EndIf
			Else ''Both closest points on the smaller side
				''One of the two extra points is (1,1,1,1)
				xsv_ext2 = xsb + 1
				ysv_ext2 = ysb + 1
				zsv_ext2 = zsb + 1
				wsv_ext2 = wsb + 1
				dx_ext2 = dx0 - 1 - 4 * SQUISH_CONSTANT_4D
				dy_ext2 = dy0 - 1 - 4 * SQUISH_CONSTANT_4D
				dz_ext2 = dz0 - 1 - 4 * SQUISH_CONSTANT_4D
				dw_ext2 = dw0 - 1 - 4 * SQUISH_CONSTANT_4D

				''Other two points are based on the shared axes.
				Dim As Byte c = cast(Byte,(aPoint And bPoint))

				if ((c and &h01) <> 0) Then
					xsv_ext0 = xsb + 2
					xsv_ext1 = xsb + 1
					dx_ext0 = dx0 - 2 - 3 * SQUISH_CONSTANT_4D
					dx_ext1 = dx0 - 1 - 3 * SQUISH_CONSTANT_4D
				Else
					xsv_ext0 = xsv_ext1 = xsb
					dx_ext0 = dx_ext1 = dx0 - 3 * SQUISH_CONSTANT_4D
				EndIf

				if ((c and &h02) <> 0) Then
					ysv_ext0 = ysv_ext1 = ysb + 1
					dy_ext0 = dy_ext1 = dy0 - 1 - 3 * SQUISH_CONSTANT_4D
					if ((c and &h01) = 0) Then
						ysv_ext0 += 1
						dy_ext0 -= 1
					Else
						ysv_ext1 += 1
						dy_ext1 -= 1
					EndIf
				Else
					ysv_ext0 = ysv_ext1 = ysb
					dy_ext0 = dy_ext1 = dy0 - 3 * SQUISH_CONSTANT_4D
				EndIf

				if ((c and &h04) <> 0) Then
					zsv_ext0 = zsv_ext1 = zsb + 1
					dz_ext0 = dz_ext1 = dz0 - 1 - 3 * SQUISH_CONSTANT_4D
					if ((c and &h03) = 0) Then
						zsv_ext0 += 1
						dz_ext0 -= 1
					Else
						zsv_ext1 += 1
						dz_ext1 -= 1
					EndIf
				Else
					zsv_ext0 = zsv_ext1 = zsb
					dz_ext0 = dz_ext1 = dz0 - 3 * SQUISH_CONSTANT_4D
				EndIf

				if ((c and &h08) <> 0) Then
					wsv_ext0 = wsb + 1
					wsv_ext1 = wsb + 2
					dw_ext0 = dw0 - 1 - 3 * SQUISH_CONSTANT_4D
					dw_ext1 = dw0 - 2 - 3 * SQUISH_CONSTANT_4D
				Else
					wsv_ext0 = wsv_ext1 = wsb
					dw_ext0 = dw_ext1 = dw0 - 3 * SQUISH_CONSTANT_4D
				EndIf
			EndIf
		Else ''One point on each "side"
			Dim As Byte c1, c2
			if (aIsBiggerSide) Then
				c1 = aPoint
				c2 = bPoint
			Else
				c1 = bPoint
				c2 = aPoint
			EndIf

			''Two contributions are the bigger-sided point with each 1 replaced with 2.
			if ((c1 and &h01) <> 0) Then
				xsv_ext0 = xsb + 2
				xsv_ext1 = xsb + 1
				dx_ext0 = dx0 - 2 - 3 * SQUISH_CONSTANT_4D
				dx_ext1 = dx0 - 1 - 3 * SQUISH_CONSTANT_4D
			Else
				xsv_ext0 = xsv_ext1 = xsb
				dx_ext0 = dx_ext1 = dx0 - 3 * SQUISH_CONSTANT_4D
			EndIf

			if ((c1 and &h02) <> 0) Then
				ysv_ext0 = ysv_ext1 = ysb + 1
				dy_ext0 = dy_ext1 = dy0 - 1 - 3 * SQUISH_CONSTANT_4D
				if ((c1 and &h01) = 0) Then
					ysv_ext0 += 1
					dy_ext0 -= 1
				Else
					ysv_ext1 += 1
					dy_ext1 -= 1
				EndIf
			Else
				ysv_ext0 = ysv_ext1 = ysb
				dy_ext0 = dy_ext1 = dy0 - 3 * SQUISH_CONSTANT_4D
			EndIf

			if ((c1 and &h04) <> 0) Then
				zsv_ext0 = zsv_ext1 = zsb + 1
				dz_ext0 = dz_ext1 = dz0 - 1 - 3 * SQUISH_CONSTANT_4D
				if ((c1 and &h03) = 0) Then
					zsv_ext0 += 1
					dz_ext0 -= 1
				Else
					zsv_ext1 += 1
					dz_ext1 -= 1
				EndIf
			Else
				zsv_ext0 = zsv_ext1 = zsb
				dz_ext0 = dz_ext1 = dz0 - 3 * SQUISH_CONSTANT_4D
			EndIf

			if ((c1 and &h08) <> 0) Then
				wsv_ext0 = wsb + 1
				wsv_ext1 = wsb + 2
				dw_ext0 = dw0 - 1 - 3 * SQUISH_CONSTANT_4D
				dw_ext1 = dw0 - 2 - 3 * SQUISH_CONSTANT_4D
			Else
				wsv_ext0 = wsv_ext1 = wsb
				dw_ext0 = dw_ext1 = dw0 - 3 * SQUISH_CONSTANT_4D
			EndIf

			''One contribution is a permutation of (1,1,1,-1) based on the smaller-sided point
			xsv_ext2 = xsb + 1
			ysv_ext2 = ysb + 1
			zsv_ext2 = zsb + 1
			wsv_ext2 = wsb + 1
			dx_ext2 = dx0 - 1 - 2 * SQUISH_CONSTANT_4D
			dy_ext2 = dy0 - 1 - 2 * SQUISH_CONSTANT_4D
			dz_ext2 = dz0 - 1 - 2 * SQUISH_CONSTANT_4D
			dw_ext2 = dw0 - 1 - 2 * SQUISH_CONSTANT_4D
			if ((c2 and &h01) = 0) Then
				xsv_ext2 -= 2
				dx_ext2 += 2
			ElseIf ((c2 and &h02) = 0) Then
				ysv_ext2 -= 2
				dy_ext2 += 2
			ElseIf ((c2 And &h04) = 0) Then
				zsv_ext2 -= 2
				dz_ext2 += 2
			else
				wsv_ext2 -= 2
				dw_ext2 += 2
			EndIf
		EndIf

		''Contribution (1,1,1,0)
		Dim As Double dx4 = dx0 - 1 - 3 * SQUISH_CONSTANT_4D
		Dim As Double dy4 = dy0 - 1 - 3 * SQUISH_CONSTANT_4D
		Dim As Double dz4 = dz0 - 1 - 3 * SQUISH_CONSTANT_4D
		Dim As Double dw4 = dw0 - 3 * SQUISH_CONSTANT_4D
		Dim As Double attn4 = 2 - dx4 * dx4 - dy4 * dy4 - dz4 * dz4 - dw4 * dw4
		if (attn4 > 0) Then
			attn4 *= attn4
			value += attn4 * attn4 * extrapolate4(xsb + 1, ysb + 1, zsb + 1, wsb + 0, dx4, dy4, dz4, dw4)
		EndIf

		''Contribution (1,1,0,1)
		Dim As Double dx3 = dx4
		Dim As Double dy3 = dy4
		Dim As Double dz3 = dz0 - 3 * SQUISH_CONSTANT_4D
		Dim As Double dw3 = dw0 - 1 - 3 * SQUISH_CONSTANT_4D
		Dim As Double attn3 = 2 - dx3 * dx3 - dy3 * dy3 - dz3 * dz3 - dw3 * dw3
		if (attn3 > 0) Then
			attn3 *= attn3
			value += attn3 * attn3 * extrapolate4(xsb + 1, ysb + 1, zsb + 0, wsb + 1, dx3, dy3, dz3, dw3)
		EndIf

		''Contribution (1,0,1,1)
		Dim As Double dx2 = dx4
		Dim As Double dy2 = dy0 - 3 * SQUISH_CONSTANT_4D
		Dim As Double dz2 = dz4
		Dim As Double dw2 = dw3
		Dim As Double attn2 = 2 - dx2 * dx2 - dy2 * dy2 - dz2 * dz2 - dw2 * dw2
		if (attn2 > 0) Then
			attn2 *= attn2
			value += attn2 * attn2 * extrapolate4(xsb + 1, ysb + 0, zsb + 1, wsb + 1, dx2, dy2, dz2, dw2)
		EndIf

		''Contribution (0,1,1,1)
		Dim As Double dx1 = dx0 - 3 * SQUISH_CONSTANT_4D
		Dim As Double dz1 = dz4
		Dim As Double dy1 = dy4
		Dim As Double dw1 = dw3
		Dim As Double attn1 = 2 - dx1 * dx1 - dy1 * dy1 - dz1 * dz1 - dw1 * dw1
		if (attn1 > 0) Then
			attn1 *= attn1
			value += attn1 * attn1 * extrapolate4(xsb + 0, ysb + 1, zsb + 1, wsb + 1, dx1, dy1, dz1, dw1)
		EndIf

		''Contribution (1,1,0,0)
		Dim As Double dx5 = dx0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dy5 = dy0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dz5 = dz0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dw5 = dw0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double attn5 = 2 - dx5 * dx5 - dy5 * dy5 - dz5 * dz5 - dw5 * dw5
		if (attn5 > 0) Then
			attn5 *= attn5
			value += attn5 * attn5 * extrapolate4(xsb + 1, ysb + 1, zsb + 0, wsb + 0, dx5, dy5, dz5, dw5)
		EndIf

		''Contribution (1,0,1,0)
		Dim As Double dx6 = dx0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dy6 = dy0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dz6 = dz0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dw6 = dw0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double attn6 = 2 - dx6 * dx6 - dy6 * dy6 - dz6 * dz6 - dw6 * dw6
		if (attn6 > 0) Then
			attn6 *= attn6
			value += attn6 * attn6 * extrapolate4(xsb + 1, ysb + 0, zsb + 1, wsb + 0, dx6, dy6, dz6, dw6)
		EndIf

		''Contribution (1,0,0,1)
		Dim As Double dx7 = dx0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dy7 = dy0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dz7 = dz0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dw7 = dw0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double attn7 = 2 - dx7 * dx7 - dy7 * dy7 - dz7 * dz7 - dw7 * dw7
		if (attn7 > 0) Then
			attn7 *= attn7
			value += attn7 * attn7 * extrapolate4(xsb + 1, ysb + 0, zsb + 0, wsb + 1, dx7, dy7, dz7, dw7)
		EndIf

		''Contribution (0,1,1,0)
		Dim As Double dx8 = dx0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dy8 = dy0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dz8 = dz0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dw8 = dw0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double attn8 = 2 - dx8 * dx8 - dy8 * dy8 - dz8 * dz8 - dw8 * dw8
		if (attn8 > 0) Then
			attn8 *= attn8
			value += attn8 * attn8 * extrapolate4(xsb + 0, ysb + 1, zsb + 1, wsb + 0, dx8, dy8, dz8, dw8)
		EndIf

		''Contribution (0,1,0,1)
		Dim As Double dx9 = dx0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dy9 = dy0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dz9 = dz0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dw9 = dw0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double attn9 = 2 - dx9 * dx9 - dy9 * dy9 - dz9 * dz9 - dw9 * dw9
		if (attn9 > 0) Then
			attn9 *= attn9
			value += attn9 * attn9 * extrapolate4(xsb + 0, ysb + 1, zsb + 0, wsb + 1, dx9, dy9, dz9, dw9)
		EndIf

		''Contribution (0,0,1,1)
		Dim As Double dx10 = dx0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dy10 = dy0 - 0 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dz10 = dz0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double dw10 = dw0 - 1 - 2 * SQUISH_CONSTANT_4D
		Dim As Double attn10 = 2 - dx10 * dx10 - dy10 * dy10 - dz10 * dz10 - dw10 * dw10
		if (attn10 > 0) Then
			attn10 *= attn10
			value += attn10 * attn10 * extrapolate4(xsb + 0, ysb + 0, zsb + 1, wsb + 1, dx10, dy10, dz10, dw10)
		EndIf
	EndIf

	''First extra vertex
	Dim As Double attn_ext0 = 2 - dx_ext0 * dx_ext0 - dy_ext0 * dy_ext0 - dz_ext0 * dz_ext0 - dw_ext0 * dw_ext0
	if (attn_ext0 > 0) then
		attn_ext0 *= attn_ext0
		value += attn_ext0 * attn_ext0 * extrapolate4(xsv_ext0, ysv_ext0, zsv_ext0, wsv_ext0, dx_ext0, dy_ext0, dz_ext0, dw_ext0)
	EndIf

	''Second extra vertex
	Dim As Double attn_ext1 = 2 - dx_ext1 * dx_ext1 - dy_ext1 * dy_ext1 - dz_ext1 * dz_ext1 - dw_ext1 * dw_ext1
	if (attn_ext1 > 0) Then
		attn_ext1 *= attn_ext1
		value += attn_ext1 * attn_ext1 * extrapolate4(xsv_ext1, ysv_ext1, zsv_ext1, wsv_ext1, dx_ext1, dy_ext1, dz_ext1, dw_ext1)
	EndIf

	''Third extra vertex
	Dim As Double attn_ext2 = 2 - dx_ext2 * dx_ext2 - dy_ext2 * dy_ext2 - dz_ext2 * dz_ext2 - dw_ext2 * dw_ext2
	if (attn_ext2 > 0) Then
		attn_ext2 *= attn_ext2
		value += attn_ext2 * attn_ext2 * extrapolate4(xsv_ext2, ysv_ext2, zsv_ext2, wsv_ext2, dx_ext2, dy_ext2, dz_ext2, dw_ext2)
	EndIf
	value/=NORM_CONSTANT_4D
	return value
End Function
