#include once "fbgfx.bi"
#include once "GL/gl.bi"
#include once "GL/glext.bi"
#Include Once "math.bi"
#ifndef NULL
#define NULL 0
#endif

type vec3
  as GLfloat x,y,z
end type


sub ErrorExit(msg as string)
  if screenptr() then screen 0
  dim as integer w,h
  screeninfo w,h : w*=0.75:h*=0.75
  screenres w,h
  print msg
  print "press any key to quit ..."
  beep : sleep : end 1
end sub

' define OpenGL proc
#define glDefine(n) dim shared as PFN##n##PROC n
' texture
glDefine(glActiveTexture)
' shader
glDefine(glCreateShader)
glDefine(glDeleteShader)
glDefine(glShaderSource)
glDefine(glCompileShader)
glDefine(glGetShaderiv)
glDefine(glGetShaderInfoLog)
' program
glDefine(glCreateProgram)
glDefine(glDeleteProgram)
glDefine(glAttachShader)
glDefine(glDetachShader)
glDefine(glLinkProgram)
glDefine(glGetProgramiv)
glDefine(glGetProgramInfoLog)
glDefine(glUseProgram)
' uniform
glDefine(glGetUniformLocation)
glDefine(glUniform1f)
glDefine(glUniform2f)
glDefine(glUniform3f)
glDefine(glUniform4f)
glDefine(glUniform1i)
#undef glDefine

sub glScreen(w as integer=1024, h as integer=768, b as integer=32, d as integer=24, s as integer=0, f as integer=0)
  if ScreenPtr() then screen 0
  ScreenControl FB.SET_GL_STENCIL_BITS,s
  ScreenControl FB.SET_GL_DEPTH_BITS  ,d
  ScreenRes w,h,b,,FB.GFX_OPENGL or iif(f<>0,FB.GFX_FULLSCREEN,0)
  Windowtitle "offline shadertoy.com"
  flip
  ' get OpenGL proc (abort if something goes wrong)
  #define glProc(n) n = ScreenGLProc(#n) : if n = 0 then ErrorExit(#n)
  ' texture
  glProc(glActiveTexture)
  ' shader
  glProc(glCreateShader)
  glProc(glDeleteShader)
  glProc(glShaderSource)
  glProc(glCompileShader)
  glProc(glGetShaderiv)
  glProc(glGetShaderInfoLog)
  ' program
  glProc(glCreateProgram)
  glProc(glDeleteProgram)
  glProc(glAttachShader)
  glProc(glDetachShader)
  glProc(glLinkProgram)
  glProc(glGetProgramiv)
  glProc(glGetProgramInfoLog)
  glProc(glUseProgram)
  ' uniform
  glProc(glGetUniformLocation)
  glProc(glUniform1f)
  glProc(glUniform2f)
  glProc(glUniform3f)
  glProc(glUniform4f)
  glProc(glUniform1i)
  #undef glProc
end sub

type ShaderToy
  declare destructor
  declare function CompileFile(Filename as string) as boolean
  declare function CompileCode(Code as string) as boolean
  as GLuint FragmentShader
  as GLuint ProgramObject
  as string Shaderlog
end type
destructor ShaderToy
  if ProgramObject then
   glUseprogram(0)
   if FragmentShader  then
     glDetachShader(ProgramObject,FragmentShader)
     glDeleteShader(FragmentShader)
   end if
   glDeleteProgram(ProgramObject)
  end if
end destructor

function ShaderToy.CompileFile(filename as string) as boolean
  dim as string code
  var hFile = FreeFile()
  if open(filename,for input, as #hFile) then
    ShaderLog = "can't read shader: " & chr(34) & filename  & chr(34) & " !"
    return false
  end if
  while not eof(hFile)
    dim as string aLine
    line input #hFile,aLine
    code &= aLine & !"\n"
  wend
  close #hFile
  return CompileCode(code)
end function

function ShaderToy.CompileCode(UserCode as string) as boolean
  dim as GLint logSize
  dim as GLint status
  dim as string FragmentProlog
	FragmentProlog & =!"#version 120\n"
  FragmentProlog & =!"uniform vec3      iResolution;  // viewport resolution (in pixels)\n"
   FragmentProlog & =!"uniform float     iTime;  // shader playback time (in seconds)\n"
   FragmentProlog & =!"uniform float     iTimeDelta;  // shader playback time (in seconds)\n"
   FragmentProlog & =!"uniform float     iGlobalTime;  // shader playback time (in seconds)\n"
   FragmentProlog & =!"uniform int     iFrame;  // shader playback time (in seconds)\n"
  FragmentProlog & =!"uniform vec4      iMouse;       // mouse pixel coords. xy: current (if MLB down), zw: click\n"
  FragmentProlog & =!"uniform vec4      iDate;        // (year, month, day, time in seconds)\n"
  FragmentProlog & =!"uniform sampler2D iChannel0;\n"
  FragmentProlog & =!"uniform sampler2D iChannel1;\n"
  FragmentProlog & =!"uniform sampler2D iChannel2;\n"
  FragmentProlog & =!"uniform sampler2D iChannel3;\n"
  dim as string FragmentEpilog
  FragmentEpilog &= !"void main() {\n"
  FragmentEpilog &= !"  vec4 color;\n"
  FragmentEpilog &= !"  // call user shader\n"
  FragmentEpilog &= !"  mainImage(color, gl_FragCoord.xy);\n"
  FragmentEpilog &= !"  color.w = 1.0;\n"
  FragmentEpilog &= !"  gl_FragColor = color;\n"
  FragmentEpilog &= !"}\n"

  dim as string FragmentCode = FragmentProlog & UserCode & FragmentEpilog
  If Usercode<>"" Then 
  	FragmentCode=UserCode' & FragmentEpilog
  endif

  FragmentShader = glCreateShader(GL_FRAGMENT_SHADER)
  if FragmentShader=0 then
    ShaderLog = "glCreateShader(GL_FRAGMENT_SHADER) failed !"
    return false
  end if
  dim as GLchar ptr pCode=strptr(FragmentCode)
  glShaderSource (FragmentShader, 1, @pCode, NULL)
  glCompileShader(FragmentShader)
  glGetShaderiv  (FragmentShader, GL_COMPILE_STATUS, @status)
  if status = GL_FALSE then
    glGetShaderiv(FragmentShader, GL_INFO_LOG_LENGTH, @logSize)
    ShaderLog = space(logSize)
    glGetShaderInfoLog(FragmentShader, logSize, NULL, cptr(GLchar ptr,strptr(ShaderLog)) )
    ShaderLog = !"glCompileShader(FragmentShader) failed !\n" & Shaderlog
    glDeleteShader(FragmentShader) : FragmentShader = 0
    return false
  end if

  ProgramObject = glCreateProgram()
  if ProgramObject=0 then
    ShaderLog = "glCreateProgram() failed !"
    glDeleteShader(FragmentShader) : FragmentShader = 0
    return false
  end if
  glAttachShader(ProgramObject,FragmentShader)
  glLinkProgram (ProgramObject)
  glGetProgramiv(ProgramObject, GL_LINK_STATUS, @status)
  if (status = GL_FALSE) then
    glGetProgramiv(ProgramObject, GL_INFO_LOG_LENGTH, @logSize)
    ShaderLog = space(logSize)
    glGetProgramInfoLog (ProgramObject, logSize, NULL, cptr(GLchar ptr,strptr(ShaderLog)) )
    ShaderLog = !"glLinkProgram() failed !\n" & Shaderlog
    glDeleteShader(FragmentShader) : FragmentShader = 0
    return false
  end if
  return true
end function



'
' main
'
' init FBScreen and OpenGL context and load GL procs.
glScreen

' get curent resolution
dim as integer scr_w,scr_h,scr_b
screeninfo scr_w,scr_h,scr_b

dim as vec3 v3
v3.x=scr_w     ' width in pixle
v3.y=scr_h     '`height in pixle
v3.z=v3.x/v3.y ' pixel ratio
' get curent resolution


dim as ShaderToy Shader

dim as string filename=command(1)

'filename = "raymarching.glsl"
'filename = "mountains.glsl"
filename = "star_shader.shader"

if len(filename)=0 then
  ' test shader in shadertoy.com style
  dim as string strCode
  strCode &= !"void mainImage(out vec4 fragColor, in vec2 fragCoord) {\n"
  strCode &= !"  vec2 uv = fragCoord.xy / iResolution.xy;\n"
  strCode &= !"  fragColor = vec4(uv,0.5+0.5*sin(iGlobalTime),1.0);\n"
  strCode &= !"}\n"
  if Shader.CompileCode(strCode)=false then
    ErrorExit Shader.ShaderLog
  end if 
else
  if Shader.CompileFile(filename)=false then
    ErrorExit Shader.ShaderLog
  end if
end if

' create or load texture
var img = ImageCreate(128,128,0,32)
if img=NULL then ErrorExit "can't create image !"

for y as integer = 0 to 127
  for x as integer = 0 to 127
    circle img,(Rnd*127,Rnd*127),Rnd*2,RGB(Rnd*256,Rnd*256,Rnd*256),,,,F
  next
next

dim as integer tw,th
dim as any ptr pixels

dim as GLint Texture0
glGenTextures(1,@Texture0)

if imageinfo(img,tw,th,,,pixels) then
  ErrorExit "can't get imageinfo image !"
end if


glBindTexture(GL_TEXTURE_2D,Texture0)

' for the cheker board I enable mipmaping
glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP   , GL_TRUE)
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR)
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR)
 
'glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
'glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
 
'glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
'glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)

' copy the pixels from FBImage to the GPU
glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,tw,th,0,GL_BGRA,GL_UNSIGNED_BYTE,pixels)
glBindTexture(GL_TEXTURE_2D,0)

' at this point the pixels from image are on the GPU memory
' so we can delete the image from CPU memory
Imagedestroy img : img=NULL

' enable shader
glUseProgram(Shader.ProgramObject)

' get uniforms locations in shader program
var iGlobalTime = glGetUniformLocation(Shader.ProgramObject,"iGlobalTime")
var iTime = glGetUniformLocation(Shader.ProgramObject,"iTime")
Var iResolution = glGetUniformLocation(Shader.ProgramObject,"iResolution")
var iMouse      = glGetUniformLocation(Shader.ProgramObject,"iMouse")
' we have 4 slots for 2D textures and / or cubemaps
var iChannel0   = glGetUniformLocation(Shader.ProgramObject, "iChannel0")
var iChannel1   = glGetUniformLocation(Shader.ProgramObject, "iChannel1")
var iChannel2   = glGetUniformLocation(Shader.ProgramObject, "iChannel2")
var iChannel3   = glGetUniformLocation(Shader.ProgramObject, "iChannel3")
var iColor1   = glGetUniformLocation(Shader.ProgramObject, "iColor1")
var iColor2   = glGetUniformLocation(Shader.ProgramObject, "iColor2")

' for testing I use only one texture in iChannel0
glUniform1i(iChannel0,0)
glActiveTexture(GL_TEXTURE0)
glBindTexture(GL_TEXTURE_2D, Texture0)

' if you use more than one texture

'glGenTextures(1,@Texture1)
'glGenTextures(1,@Texture2)
'glGenTextures(1,@Texture3)

'glUniform1i(iChannel1,1)
'glActiveTexture(GL_TEXTURE0)
'glBindTexture(GL_TEXTURE_2D, Texture1)

'glUniform1i(iChannel2,2)
'glActiveTexture(GL_TEXTURE0)
'glBindTexture(GL_TEXTURE_2D, Texture2)

'glUniform1i(iChannel3,3)
'glActiveTexture(GL_TEXTURE0)
'glBindTexture(GL_TEXTURE_2D, Texture3)
dim as integer ox,oy,ob,mx,my,mb,frames,fps
dim as double tRuntime,tStart = Timer()
dim as double tNow,tLast = tStart
Dim tempera As Double=0
' set vec3 iResolution
glUniform3f(iResolution,v3.x,v3.y,v3.z)
' set uniform float iGlobalTime = 0
glUniform1f(iGlobalTime,tRuntime)

'dim as integer mx,my,mb,frames,fps
'dim as double tStart = Timer()
'dim as double tLast=tStart
while inkey()=""
	tempera+=1
  ' draw a rectangle (2 triangles over the whole screen)
  ' this trigers the shader code for every pixel
  glRectf(-1,-1,1,1)
  ' swap the buffer on screen
  flip
  ' count the frames
  frames+=1

  ' set uniform vec4 iMouse
  if getMouse(mx,my,,mb)=0 then
    ' update only if one or more properties are changed
    if mx<>ox orelse my=oy orelse mb<>ob then
      ox=mx:oy=my:ob=mb
      if mb<>1 then mb=0 ' triger only the left mouse button
      glUniform4f(iMouse,mx,scr_h-my,mb,mb)
    end if
  end If
  Dim t1 As UInteger=Temperature2RGB(tempera)
  Dim tr1 As Single=((t1 Shr 16)And &hff) /256
  Dim tg1 As Single=((t1 Shr 8)And &hff) /256
  Dim tb1 As Single=((t1 )And &hff) /256
  Dim t2 As uInteger=Temperature2RGB(tempera-(tempera/1000))
  Dim tr2 As Single=((t2 Shr 16)And &hff) /256
  Dim tg2 As Single=((t2 Shr 8)And &hff) /256
  Dim tb2 As Single=((t2 )And &hff) /256
  'Print tempera,t1,tr1,tg1,tb1
      glUniform3f(iColor1,tr1,tg1,tb1)
      glUniform3f(iColor2,tr2,tg2,tb2)

  ' get the time
  tNow=Timer() : tRuntime=tNow-tStart
  ' update uniform float iGlobalTime
  glUniform1f(iGlobalTime,tRuntime)
  glUniform1f(iTime,tRuntime)
 
  ' update fps (not every frame Windowtitle on Windows or Linux are slow)
  if frames mod 1=0 then
    fps=100/(tNow-tLast) : tLast=tNow
    Windowtitle "" & scr_w & " x " & scr_h & " x "  & scr_b & " fps: " & fps
  end if
wend

' disable shader
glUseProgram(0)