/'
   auxiliary (read: rushed) functions for performing some drawings on the
   3D plane
'/
function clipPoint( byref cam as cameras, byref s1 as vec4) as integer
	/'
      clip both vertices of a line if necessary

      note that this 'clipping' is only done on the near plane, not on the scene

      the sensible (and correct) way to do this would be to clip them against
      the view frustum; but alas, in order to keep the code short and clear
      the calculation of the frustum was removed from the camera class. Hence
      this crappy function.

      Clipping (especially against the near plane) is important, because if you
      don't clip, when one of the line endpoints goes behind the near clipping
      plane, it gets a negative coordinate (due to the projection), and the line
      segment is no longer valid
   '/
	if( s1.z >= cam.nearClip())  Then
		return( -1 )
		'elseIf( s1.z < cam.nearClip() ) then
		'    ' First coordinate behind the camera, clip it
		'    s1.x = s1.x + ((cam.nearClip() - s1.z))
		'    s1.y = s1.y + ((cam.nearClip() - s1.z))
		'    s1.z = cam.nearClip()
		'
		'    return( -1 )
	else
		' coordinates behind the camera, don't draw
		return( 0 )
	end if
end Function
function clipSegment( byref cam as cameras, byref s1 as vec4, byref s2 as vec4 ) as integer
	/'
      clip both vertices of a line if necessary

      note that this 'clipping' is only done on the near plane, not on the scene

      the sensible (and correct) way to do this would be to clip them against
      the view frustum; but alas, in order to keep the code short and clear
      the calculation of the frustum was removed from the camera class. Hence
      this crappy function.

      Clipping (especially against the near plane) is important, because if you
      don't clip, when one of the line endpoints goes behind the near clipping
      plane, it gets a negative coordinate (due to the projection), and the line
      segment is no longer valid
   '/
	if( s1.z >= cam.nearClip() and s2.z >= cam.nearClip() ) then
		' Neither one is behind the camera, draw them both
		return( -1 )
	elseIf( s1.z < cam.nearClip() and s2.z >= cam.nearClip() ) then
		' First coordinate behind the camera, clip it
		s1.x = s1.x + ((cam.nearClip() - s1.z) / (s2.z - s1.z)) * (s2.x - s1.x)
		s1.y = s1.y + ((cam.nearClip() - s1.z) / (s2.z - s1.z)) * (s2.y - s1.y)
		s1.z = cam.nearClip()

		return( -1 )
	elseIf( s1.z >= cam.nearClip() and s2.z < cam.nearClip() ) then
		' Second coordinate behind the camera, clip it
		s2.x = s1.x + ((cam.nearClip() - s1.z) / (s2.z - s1.z)) * (s2.x - s1.x)
		s2.y = s1.y + ((cam.nearClip() - s1.z) / (s2.z - s1.z)) * (s2.y - s1.y)
		s2.z = cam.nearClip()

		return( -1 )
	else
		' Both coordinates behind the camera, don't draw
		return( 0 )
	end if
end function

sub drawLine3d( byref cam as cameras, byref p1 As vec4, byref p2 As vec4, byval c as uint32,p As UShort=&hffff )
	dim as vec4 pos1, pos2

	pos1 = cam.transform( p1 - cam.getPos() )
	pos2 = cam.transform( p2 - cam.getPos() )

	dim as double x1, y1, x2, y2

	dim as integer visible = clipSegment( cam, pos1, pos2 )

	if( visible ) then
		dim as double z1, z2

		cam.perspective( pos1, cam.projectionPlane, x1, y1, z1 )
		cam.perspective( pos2, cam.projectionPlane, x2, y2, z2 )

		line( x1, y1 ) - ( x2, y2 ), c,,p

	end if
end sub
Sub drawCircle3d(ByRef cam As cameras,ByRef p As vec4, ByRef r As Double,e As Double, ByVal nr_s As uint32, ByVal c As uint32)
	Dim As vec4 p1,p2
	For i As UInteger=0 To 360 Step nr_s
		p1.x=p.x+Sin(i*degToRad)*r
		p1.y=p.y+Sin(i*degToRad)*e
		p1.z=p.z+Cos(i*degToRad)*r
		p2.x=p.x+Sin((i+nr_s)*degToRad)*r
		p2.y=p.y+Sin(i*degToRad)*e
		p2.z=p.z+Cos((i+nr_s)*degToRad)*r
		drawLine3d(cam,p1,p2,c)
	Next
End Sub
Sub drawPoint3d(ByRef cam As cameras,ByRef p As vec4,ByVal r As double, ByVal c As uint32)
	dim as vec4 pos1

	pos1 = cam.transform( p - cam.getPos() )

	dim as double x1, y1, x2, y2,r1

	dim as integer visible = clipPoint( cam, pos1)
	r1=r/abs(pos1.z)
	If( visible ) Then
		If r1>=0.5 Then
			dim as double z1, z2
			cam.perspective( pos1, cam.projectionPlane, x1, y1, z1 )
			circle( x1, y1 ), r1, c,,,,F

	EndIf
	endif
End Sub
Sub drawPixel3d(ByRef cam As cameras,ByRef p As vec4,ByVal r As double, ByVal c As uint32)
	dim as vec4 pos1

	pos1 = cam.transform( p - cam.getPos() )

	dim as double x1, y1, x2, y2,r1

	dim as integer visible = clipPoint( cam, pos1)
	'r1=r/abs(pos1.z)
	If( visible ) Then
			dim as double z1, z2
			cam.perspective( pos1, cam.projectionPlane, x1, y1, z1 )
			If x1>0 And x1<cam.projectionPlane.width And y1>0 And y1<cam.projectionPlane.height Then
			'circle( x1, y1 ), r, c,,,,F
			Circle( x1+cam.projectionPlane.x, y1+cam.projectionPlane.y ), r, c,,,,F
			EndIf
endif
End Sub
Sub drawBox3d(ByRef cam As cameras,ByRef p As vec4,ByVal r As double, ByVal c As uint32)
	dim as vec4 pos1

	pos1 = cam.transform( p - cam.getPos() )

	dim as double x1, y1, x2, y2,r1

	dim as integer visible = clipPoint( cam, pos1)
	r1=r/abs(pos1.z)
	If( visible ) Then
		If r1>=0.5 Then
			dim as double z1, z2
			cam.perspective( pos1, cam.projectionPlane, x1, y1, z1 )
			cam.perspective( pos1, cam.projectionPlane, x2, y2, z2 )
		line( x1-r1, y1-r1 ) - ( x2+r1, y2+r1 ), c,B

	EndIf
	endif
End Sub
Sub drawText3d(ByRef cam As cameras,ByRef p As vec4,ByVal t As String,ByVal c As uint32,size As Double=8)
	dim as vec4 pos1

	pos1 = cam.transform( p - cam.getPos() )

	dim as double x1, y1, z1

	dim as integer visible = clipPoint( cam, pos1)
	If( visible ) Then
		cam.perspective( pos1, cam.projectionPlane, x1, y1, z1 )
		'print_text(x1,y1,t,font1,size,c)
		Draw string( x1, y1 ), t, c
	EndIf
End Sub
Sub drawText3dshadowed(ByRef cam As cameras,ByRef p As vec4,ByVal t As String,ByVal c As uint32, ByVal c2 As uint32,size As Double=8)
	dim as vec4 pos1

	pos1 = cam.transform( p - cam.getPos() )

	dim as double x1, y1, z1

	dim as integer visible = clipPoint( cam, pos1)
	If( visible ) Then
		cam.perspective( pos1, cam.projectionPlane, x1, y1, z1 )
		'print_text(x1+1,y1+1,t,font1,size,c2)
		'print_text(x1,y1,t,font1,size,c)
		Draw String(x1+1,y1+1),t,c2
		Draw string( x1, y1 ), t, c
	EndIf
End Sub
Function MouseOver(ByRef cam As cameras,ByRef p As vec4,ByVal r As Double ) As boolean
	dim as vec4 pos1

	pos1 = cam.transform( p - cam.getPos() )

	dim as double x1, y1, z1,r1
	r1=r/abs(pos1.z)

	dim as integer visible = clipPoint( cam, pos1)
	If( visible ) Then
		If r1>=1.5 Then
			cam.perspective( pos1, cam.projectionPlane, x1, y1, z1 )
			If ((MOUSEX<x1+r1)And (MOUSEX>x1-r1)) And ((MOUSEY<y1+r1)And (MOUSEY>y1-r1)) Then Return TRUE
		EndIf
	EndIf
	Return FALSE
End Function
